#
#    Project: SILX
#             https://github.com/silx-kit/silx
#
#    Copyright (C) 2012-2019 European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

"""A module for performing basic statistical analysis (min, max, mean, std) on
large data where numpy is not very efficient.
"""

__author__ = "Jerome Kieffer"
__license__ = "MIT"
__date__ = "19/05/2021"
__copyright__ = "2012-2019, ESRF, Grenoble"
__contact__ = "jerome.kieffer@esrf.fr"

import logging
import numpy
from collections import OrderedDict, namedtuple
from math import sqrt

from .common import pyopencl
from .processing import EventDescription, OpenclProcessing, BufferDescription
from .utils import concatenate_cl_kernel

if pyopencl:
    mf = pyopencl.mem_flags
    from pyopencl.reduction import ReductionKernel
    try:
        from pyopencl import cltypes
    except ImportError:
        v = pyopencl.array.vec()
        float8 = v.float8
    else:
        float8 = cltypes.float8

else:
    raise ImportError("pyopencl is not installed")
logger = logging.getLogger(__name__)

StatResults = namedtuple("StatResults", ["min", "max", "cnt", "sum", "mean",
                                         "var", "std"])
zero8 = "(float8)(FLT_MAX, -FLT_MAX, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f)"
#                    min      max    cnt  cnt_e  sum   sum_e  var  var_e


class Statistics(OpenclProcessing):
    """A class for doing statistical analysis using OpenCL

    :param List[int] size: Shape of input data to treat
    :param numpy.dtype dtype: Input data type
    :param numpy.ndarray template: Data template to extract size & dtype
    :param ctx: Actual working context, left to None for automatic
                initialization from device type or platformid/deviceid
    :param str devicetype: Type of device, can be "CPU", "GPU", "ACC" or "ALL"
    :param int platformid: Platform identifier as given by clinfo
    :param int deviceid: Device identifier as given by clinfo
    :param int block_size:
        Preferred workgroup size, may vary depending on the outcome of the compilation
    :param bool profile:
        Switch on profiling to be able to profile at the kernel level,
        store profiling elements (makes code slightly slower)
    """
    buffers = [
        BufferDescription("raw", 1, numpy.float32, mf.READ_ONLY),
        BufferDescription("converted", 1, numpy.float32, mf.READ_WRITE),
    ]
    kernel_files = ["preprocess.cl"]
    mapping = {numpy.int8: "s8_to_float",
               numpy.uint8: "u8_to_float",
               numpy.int16: "s16_to_float",
               numpy.uint16: "u16_to_float",
               numpy.uint32: "u32_to_float",
               numpy.int32: "s32_to_float"}

    def __init__(self, size=None, dtype=None, template=None,
                 ctx=None, devicetype="all", platformid=None, deviceid=None,
                 block_size=None, profile=False
                 ):
        OpenclProcessing.__init__(self, ctx=ctx, devicetype=devicetype,
                                  platformid=platformid, deviceid=deviceid,
                                  block_size=block_size, profile=profile)
        self.size = size
        self.dtype = dtype
        if template is not None:
            self.size = template.size
            self.dtype = template.dtype

        self.buffers = [BufferDescription(i.name, i.size * self.size, i.dtype, i.flags)
                        for i in self.__class__.buffers]

        self.allocate_buffers(use_array=True)
        self.compile_kernels()
        self.set_kernel_arguments()

    def set_kernel_arguments(self):
        """Parametrize all kernel arguments"""
        for val in self.mapping.values():
            self.cl_kernel_args[val] = OrderedDict(((i, self.cl_mem[i]) for i in ("raw", "converted")))

    def compile_kernels(self):
        """Compile the kernel"""
        OpenclProcessing.compile_kernels(self,
                                         self.kernel_files,
                                         "-D NIMAGE=%i" % self.size)
        compiler_options = self.get_compiler_options(x87_volatile=True)
        src = concatenate_cl_kernel(("doubleword.cl", "statistics.cl"))
        self.reduction_comp = ReductionKernel(self.ctx,
                                              dtype_out=float8,
                                              neutral=zero8,
                                              map_expr="map_statistics(data, i)",
                                              reduce_expr="reduce_statistics(a,b)",
                                              arguments="__global float *data",
                                              preamble=src,
                                              options=compiler_options)
        self.reduction_simple = ReductionKernel(self.ctx,
                                                dtype_out=float8,
                                                neutral=zero8,
                                                map_expr="map_statistics(data, i)",
                                                reduce_expr="reduce_statistics_simple(a,b)",
                                                arguments="__global float *data",
                                                preamble=src,
                                                options=compiler_options)

        if "cl_khr_fp64" in self.device.extensions:
            self.reduction_double = ReductionKernel(self.ctx,
                                                    dtype_out=float8,
                                                    neutral=zero8,
                                                    map_expr="map_statistics(data, i)",
                                                    reduce_expr="reduce_statistics_double(a,b)",
                                                    arguments="__global float *data",
                                                    preamble=src,
                                                    options=compiler_options)
        else:
            logger.info("Device %s does not support double-precision arithmetics, fall-back on compensated one", self.device)
            self.reduction_double = self.reduction_comp

    def send_buffer(self, data, dest):
        """
        Send a numpy array to the device, including the cast on the device if
        possible

        :param numpy.ndarray data: numpy array with data
        :param dest: name of the buffer as registered in the class
        """
        logger.info("send data to %s", dest)
        dest_type = numpy.dtype([i.dtype for i in self.buffers if i.name == dest][0])
        events = []
        if (data.dtype == dest_type) or (data.dtype.itemsize > dest_type.itemsize):
            copy_image = pyopencl.enqueue_copy(self.queue,
                                               self.cl_mem[dest].data,
                                               numpy.ascontiguousarray(data, dest_type))
            events.append(EventDescription("copy H->D %s" % dest, copy_image))
        else:
            copy_image = pyopencl.enqueue_copy(self.queue,
                                               self.cl_mem["raw"].data,
                                               numpy.ascontiguousarray(data))
            kernel = getattr(self.program, self.mapping[data.dtype.type])
            cast_to_float = kernel(self.queue,
                                   (self.size,),
                                   None,
                                   self.cl_mem["raw"].data,
                                   self.cl_mem[dest].data)
            events += [
                EventDescription("copy H->D raw", copy_image),
                EventDescription(f"cast to float {dest}", cast_to_float)
            ]
        if self.profile:
            self.events += events
        return events

    def process(self, data, comp=True):
        """Actually calculate the statics on the data

        :param numpy.ndarray data: numpy array with the image
        :param comp: use Kahan compensated arithmetics for the calculation 
        :return: Statistics named tuple
        :rtype: StatResults
        """
        if data.ndim != 1:
            data = data.ravel()
        size = data.size
        assert size <= self.size, "size is OK"
        events = []
        if comp is True:
            comp = "comp"
        elif comp is False:
            comp = "single"
        else:
            comp = comp.lower()
        with self.sem:
            self.send_buffer(data, "converted")
            if comp in ("single", "fp32", "float32"):
                reduction = self.reduction_simple
            elif comp in ("double", "fp64", "float64"):
                reduction = self.reduction_double
            else:
                reduction = self.reduction_comp
            res_d, evt = reduction(self.cl_mem["converted"][:self.size],
                                   queue=self.queue,
                                   return_event=True)
            events.append(EventDescription(f"statistical reduction {comp}", evt))
            if self.profile:
                self.events += events
            res_h = res_d.get()
        min_ = 1.0 * res_h["s0"]
        max_ = 1.0 * res_h["s1"]
        count = 1.0 * res_h["s2"] + res_h["s3"]
        sum_ = 1.0 * res_h["s4"] + res_h["s5"]
        m2 = 1.0 * res_h["s6"] + res_h["s7"]
        var = m2 / (count - 1.0)
        res = StatResults(min_,
                          max_,
                          count,
                          sum_,
                          sum_ / count,
                          var,
                          sqrt(var))
        return res

    __call__ = process



# /*##########################################################################
#
# Copyright (c) 2017-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module contains utils class for axes management.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "20/11/2018"

import functools
import logging
from contextlib import contextmanager
import weakref
import silx.utils.weakref as silxWeakref
from silx.gui.plot.items.axis import Axis, XAxis, YAxis
from ...qt.inspect import isValid as _isQObjectValid


_logger = logging.getLogger(__name__)


class SyncAxes(object):
    """Synchronize a set of plot axes together.

    It is created with the expected axes and starts to synchronize them.

    It can be customized to synchronize limits, scale, and direction of axes
    together. By default everything is synchronized.

    The API :meth:`start` and :meth:`stop` can be used to enable/disable the
    synchronization while this object is still alive.

    If this object is destroyed the synchronization stop.

    .. versionadded:: 0.6
    """

    def __init__(self, axes,
                 syncLimits=True,
                 syncScale=True,
                 syncDirection=True,
                 syncCenter=False,
                 syncZoom=False,
                 filterHiddenPlots=False
                 ):
        """
        Constructor

        :param list(Axis) axes: A list of axes to synchronize together
        :param bool syncLimits: Synchronize axes limits
        :param bool syncScale: Synchronize axes scale
        :param bool syncDirection: Synchronize axes direction
        :param bool syncCenter: Synchronize the center of the axes in the center
            of the plots
        :param bool syncZoom: Synchronize the zoom of the plot
        :param bool filterHiddenPlots: True to avoid updating hidden plots.
            Default: False.
        """
        object.__init__(self)

        def implies(x, y): return bool(y ** x)

        assert(implies(syncZoom, not syncLimits))
        assert(implies(syncCenter, not syncLimits))
        assert(implies(syncLimits, not syncCenter))
        assert(implies(syncLimits, not syncZoom))

        self.__filterHiddenPlots = filterHiddenPlots
        self.__locked = False
        self.__axisRefs = []
        self.__syncLimits = syncLimits
        self.__syncScale = syncScale
        self.__syncDirection = syncDirection
        self.__syncCenter = syncCenter
        self.__syncZoom = syncZoom
        self.__callbacks = None
        self.__lastMainAxis = None

        for axis in axes:
            self.addAxis(axis)

        self.start()

    def start(self):
        """Start synchronizing axes together.

        The first axis is used as the reference for the first synchronization.
        After that, any changes to any axes will be used to synchronize other
        axes.
        """
        if self.isSynchronizing():
            raise RuntimeError("Axes already synchronized")
        self.__callbacks = {}

        axes = self.__getAxes()

        # register callback for further sync
        for axis in axes:
            self.__connectAxes(axis)
        self.synchronize()

    def isSynchronizing(self):
        """Returns true if events are connected to the axes to synchronize them
        all together

        :rtype: bool
        """
        return self.__callbacks is not None

    def __connectAxes(self, axis):
        refAxis = weakref.ref(axis)
        callbacks = []
        if self.__syncLimits:
            # the weakref is needed to be able ignore self references
            callback = silxWeakref.WeakMethodProxy(self.__axisLimitsChanged)
            callback = functools.partial(callback, refAxis)
            sig = axis.sigLimitsChanged
            sig.connect(callback)
            callbacks.append(("sigLimitsChanged", callback))
        elif self.__syncCenter and self.__syncZoom:
            # the weakref is needed to be able ignore self references
            callback = silxWeakref.WeakMethodProxy(self.__axisCenterAndZoomChanged)
            callback = functools.partial(callback, refAxis)
            sig = axis.sigLimitsChanged
            sig.connect(callback)
            callbacks.append(("sigLimitsChanged", callback))
        elif self.__syncZoom:
            raise NotImplementedError()
        elif self.__syncCenter:
            # the weakref is needed to be able ignore self references
            callback = silxWeakref.WeakMethodProxy(self.__axisCenterChanged)
            callback = functools.partial(callback, refAxis)
            sig = axis.sigLimitsChanged
            sig.connect(callback)
            callbacks.append(("sigLimitsChanged", callback))
        if self.__syncScale:
            # the weakref is needed to be able ignore self references
            callback = silxWeakref.WeakMethodProxy(self.__axisScaleChanged)
            callback = functools.partial(callback, refAxis)
            sig = axis.sigScaleChanged
            sig.connect(callback)
            callbacks.append(("sigScaleChanged", callback))
        if self.__syncDirection:
            # the weakref is needed to be able ignore self references
            callback = silxWeakref.WeakMethodProxy(self.__axisInvertedChanged)
            callback = functools.partial(callback, refAxis)
            sig = axis.sigInvertedChanged
            sig.connect(callback)
            callbacks.append(("sigInvertedChanged", callback))

        if self.__filterHiddenPlots:
            # the weakref is needed to be able ignore self references
            callback = silxWeakref.WeakMethodProxy(self.__axisVisibilityChanged)
            callback = functools.partial(callback, refAxis)
            plot = axis._getPlot()
            plot.sigVisibilityChanged.connect(callback)
            callbacks.append(("sigVisibilityChanged", callback))

        self.__callbacks[refAxis] = callbacks

    def __disconnectAxes(self, axis):
        if axis is not None and _isQObjectValid(axis):
            ref = weakref.ref(axis)
            callbacks = self.__callbacks.pop(ref)
            for sigName, callback in callbacks:
                if sigName == "sigVisibilityChanged":
                    obj = axis._getPlot()
                else:
                    obj = axis
                if obj is not None:
                    sig = getattr(obj, sigName)
                    sig.disconnect(callback)

    def addAxis(self, axis):
        """Add a new axes to synchronize.

        :param ~silx.gui.plot.items.Axis axis: The axis to synchronize
        """
        self.__axisRefs.append(weakref.ref(axis))
        if self.isSynchronizing():
            self.__connectAxes(axis)
            # This could be done faster as only this axis have to be fixed
            self.synchronize()

    def removeAxis(self, axis):
        """Remove an axis from the synchronized axes.

        :param ~silx.gui.plot.items.Axis axis: The axis to remove
        """
        ref = weakref.ref(axis)
        self.__axisRefs.remove(ref)
        if self.isSynchronizing():
            self.__disconnectAxes(axis)

    def synchronize(self, mainAxis=None):
        """Synchronize programatically all the axes.

        :param ~silx.gui.plot.items.Axis mainAxis:
            The axis to take as reference (Default: the first axis).
        """
        # sync the current state
        axes = self.__getAxes()
        if len(axes) == 0:
            return

        if mainAxis is None:
            mainAxis = axes[0]

        refMainAxis = weakref.ref(mainAxis)
        if self.__syncLimits:
            self.__axisLimitsChanged(refMainAxis, *mainAxis.getLimits())
        elif self.__syncCenter and self.__syncZoom:
            self.__axisCenterAndZoomChanged(refMainAxis, *mainAxis.getLimits())
        elif self.__syncCenter:
            self.__axisCenterChanged(refMainAxis, *mainAxis.getLimits())
        if self.__syncScale:
            self.__axisScaleChanged(refMainAxis, mainAxis.getScale())
        if self.__syncDirection:
            self.__axisInvertedChanged(refMainAxis, mainAxis.isInverted())

    def stop(self):
        """Stop the synchronization of the axes"""
        if not self.isSynchronizing():
            raise RuntimeError("Axes not synchronized")
        for ref in list(self.__callbacks.keys()):
            axis = ref()
            self.__disconnectAxes(axis)
        self.__callbacks = None

    def __del__(self):
        """Destructor"""
        # clean up references
        if self.__callbacks is not None:
            self.stop()

    def __getAxes(self):
        """Returns list of existing axes.

        :rtype: List[Axis]
        """
        axes = [ref() for ref in self.__axisRefs]
        return [axis for axis in axes if axis is not None]

    @contextmanager
    def __inhibitSignals(self):
        self.__locked = True
        yield
        self.__locked = False

    def __axesToUpdate(self, changedAxis):
        for axis in self.__getAxes():
            if axis is changedAxis:
                continue
            if self.__filterHiddenPlots:
                plot = axis._getPlot()
                if not plot.isVisible():
                    continue
            yield axis

    def __axisVisibilityChanged(self, changedAxis, isVisible):
        if not isVisible:
            return
        if self.__locked:
            return
        changedAxis = changedAxis()
        if self.__lastMainAxis is None:
            self.__lastMainAxis = self.__axisRefs[0]
        mainAxis = self.__lastMainAxis
        mainAxis = mainAxis()
        self.synchronize(mainAxis=mainAxis)
        # force back the main axis
        self.__lastMainAxis = weakref.ref(mainAxis)

    def __getAxesCenter(self, axis, vmin, vmax):
        """Returns the value displayed in the center of this axis range.

        :rtype: float
        """
        scale = axis.getScale()
        if scale == Axis.LINEAR:
            center = (vmin + vmax) * 0.5
        else:
            raise NotImplementedError("Log scale not implemented")
        return center

    def __getRangeInPixel(self, axis):
        """Returns the size of the axis in pixel"""
        bounds = axis._getPlot().getPlotBoundsInPixels()
        # bounds: left, top, width, height
        if isinstance(axis, XAxis):
            return bounds[2]
        elif isinstance(axis, YAxis):
            return bounds[3]
        else:
            assert(False)

    def __getLimitsFromCenter(self, axis, pos, pixelSize=None):
        """Returns the limits to apply to this axis to move the `pos` into the
        center of this axis.

        :param Axis axis:
        :param float pos: Position in the center of the computed limits
        :param Union[None,float] pixelSize: Pixel size to apply to compute the
            limits. If `None` the current pixel size is applyed.
        """
        scale = axis.getScale()
        if scale == Axis.LINEAR:
            if pixelSize is None:
                # Use the current pixel size of the axis
                limits = axis.getLimits()
                valueRange = limits[0] - limits[1]
                a = pos - valueRange * 0.5
                b = pos + valueRange * 0.5
            else:
                pixelRange = self.__getRangeInPixel(axis)
                a = pos - pixelRange * 0.5 * pixelSize
                b = pos + pixelRange * 0.5 * pixelSize

        else:
            raise NotImplementedError("Log scale not implemented")
        if a > b:
            return b, a
        return a, b

    def __axisLimitsChanged(self, changedAxis, vmin, vmax):
        if self.__locked:
            return
        self.__lastMainAxis = changedAxis
        changedAxis = changedAxis()
        with self.__inhibitSignals():
            for axis in self.__axesToUpdate(changedAxis):
                axis.setLimits(vmin, vmax)

    def __axisCenterAndZoomChanged(self, changedAxis, vmin, vmax):
        if self.__locked:
            return
        self.__lastMainAxis = changedAxis
        changedAxis = changedAxis()
        with self.__inhibitSignals():
            center = self.__getAxesCenter(changedAxis, vmin, vmax)
            pixelRange = self.__getRangeInPixel(changedAxis)
            if pixelRange == 0:
                return
            pixelSize = (vmax - vmin) / pixelRange
            for axis in self.__axesToUpdate(changedAxis):
                vmin, vmax = self.__getLimitsFromCenter(axis, center, pixelSize)
                axis.setLimits(vmin, vmax)

    def __axisCenterChanged(self, changedAxis, vmin, vmax):
        if self.__locked:
            return
        self.__lastMainAxis = changedAxis
        changedAxis = changedAxis()
        with self.__inhibitSignals():
            center = self.__getAxesCenter(changedAxis, vmin, vmax)
            for axis in self.__axesToUpdate(changedAxis):
                vmin, vmax = self.__getLimitsFromCenter(axis, center)
                axis.setLimits(vmin, vmax)

    def __axisScaleChanged(self, changedAxis, scale):
        if self.__locked:
            return
        self.__lastMainAxis = changedAxis
        changedAxis = changedAxis()
        with self.__inhibitSignals():
            for axis in self.__axesToUpdate(changedAxis):
                axis.setScale(scale)

    def __axisInvertedChanged(self, changedAxis, isInverted):
        if self.__locked:
            return
        self.__lastMainAxis = changedAxis
        changedAxis = changedAxis()
        with self.__inhibitSignals():
            for axis in self.__axesToUpdate(changedAxis):
                axis.setInverted(isInverted)



# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Module testing silx.app.view"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "07/06/2018"


import os
import shutil
import sys
import tempfile
import unittest
import logging
import subprocess
import pytest

from .. import main
from silx import __main__ as silx_main

_logger = logging.getLogger(__name__)


@pytest.mark.usefixtures("qapp")
class TestLauncher(unittest.TestCase):
    """Test command line parsing"""

    def testHelp(self):
        # option -h must cause a raise SystemExit or a return 0
        try:
            parser = main.createParser()
            parser.parse_args(["view", "--help"])
            result = 0
        except SystemExit as e:
            result = e.args[0]
        self.assertEqual(result, 0)

    def testWrongOption(self):
        try:
            parser = main.createParser()
            parser.parse_args(["view", "--foo"])
            self.fail()
        except SystemExit as e:
            result = e.args[0]
        self.assertNotEqual(result, 0)

    def testWrongFile(self):
        try:
            parser = main.createParser()
            result = parser.parse_args(["view", "__file.not.found__"])
            result = 0
        except SystemExit as e:
            result = e.args[0]
        self.assertEqual(result, 0)

    def executeAsScript(self, filename, *args):
        """Execute a command line.

        Log output as debug in case of bad return code.
        """
        env = self.createTestEnv()

        with tempfile.TemporaryDirectory() as tmpdir:
            # Copy file to temporary dir to avoid import from current dir.
            script = os.path.join(tmpdir, 'launcher.py')
            shutil.copyfile(filename, script)
            command_line = [sys.executable, script] + list(args)

            _logger.info("Execute: %s", " ".join(command_line))
            p = subprocess.Popen(command_line,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 env=env)
            out, err = p.communicate()
            _logger.info("Return code: %d", p.returncode)
            try:
                out = out.decode('utf-8')
            except UnicodeError:
                pass
            try:
                err = err.decode('utf-8')
            except UnicodeError:
                pass

            if p.returncode != 0:
                _logger.info("stdout:")
                _logger.info("%s", out)
                _logger.info("stderr:")
                _logger.info("%s", err)
            else:
                _logger.debug("stdout:")
                _logger.debug("%s", out)
                _logger.debug("stderr:")
                _logger.debug("%s", err)
            self.assertEqual(p.returncode, 0)

    def createTestEnv(self):
        """
        Returns an associated environment with a working project.
        """
        env = dict((str(k), str(v)) for k, v in os.environ.items())
        env["PYTHONPATH"] = os.pathsep.join(sys.path)
        return env

    def testExecuteViewHelp(self):
        """Test if the main module is well connected.

        Uses subprocess to avoid to parasite the current environment.
        """
        self.executeAsScript(main.__file__, "--help")

    def testExecuteSilxViewHelp(self):
        """Test if the main module is well connected.

        Uses subprocess to avoid to parasite the current environment.
        """
        self.executeAsScript(silx_main.__file__, "view", "--help")



# /*##########################################################################
#
# Copyright (c) 2016-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "19/02/2019"

import os
import tempfile
import pytest
from contextlib import contextmanager

import numpy
from ..DataViewer import DataViewer
from ..DataViews import DataView
from .. import DataViews

from silx.gui import qt

from silx.gui.data.DataViewerFrame import DataViewerFrame
from silx.gui.utils.testutils import SignalListener
from silx.gui.utils.testutils import TestCaseQt

import h5py


class _DataViewMock(DataView):
    """Dummy view to display nothing"""

    def __init__(self, parent):
        DataView.__init__(self, parent)

    def axesNames(self, data, info):
        return []

    def createWidget(self, parent):
        return qt.QLabel(parent)

    def getDataPriority(self, data, info):
        return 0


class _TestAbstractDataViewer(TestCaseQt):
    __test__ = False  # ignore abstract class

    def create_widget(self):
        # Avoid to raise an error when testing the full module
        self.skipTest("Not implemented")

    @contextmanager
    def h5_temporary_file(self):
        # create tmp file
        fd, tmp_name = tempfile.mkstemp(suffix=".h5")
        os.close(fd)
        data = numpy.arange(3 * 3 * 3)
        data.shape = 3, 3, 3
        # create h5 data
        h5file = h5py.File(tmp_name, "w")
        h5file["data"] = data
        yield h5file
        # clean up
        h5file.close()
        os.unlink(tmp_name)

    def test_text_data(self):
        data_list = ["aaa", int, 8, self]
        widget = self.create_widget()
        for data in data_list:
            widget.setData(data)
            self.assertEqual(DataViews.RAW_MODE, widget.displayMode())

    def test_plot_1d_data(self):
        data = numpy.arange(3 ** 1)
        data.shape = [3] * 1
        widget = self.create_widget()
        widget.setData(data)
        availableModes = set([v.modeId() for v in widget.currentAvailableViews()])
        self.assertEqual(DataViews.RAW_MODE, widget.displayMode())
        self.assertIn(DataViews.PLOT1D_MODE, availableModes)

    def test_image_data(self):
        data = numpy.arange(3 ** 2)
        data.shape = [3] * 2
        widget = self.create_widget()
        widget.setData(data)
        availableModes = set([v.modeId() for v in widget.currentAvailableViews()])
        self.assertEqual(DataViews.RAW_MODE, widget.displayMode())
        self.assertIn(DataViews.IMAGE_MODE, availableModes)

    def test_image_bool(self):
        data = numpy.zeros((10, 10), dtype=bool)
        data[::2, ::2] = True
        widget = self.create_widget()
        widget.setData(data)
        availableModes = set([v.modeId() for v in widget.currentAvailableViews()])
        self.assertEqual(DataViews.RAW_MODE, widget.displayMode())
        self.assertIn(DataViews.IMAGE_MODE, availableModes)

    def test_image_complex_data(self):
        data = numpy.arange(3 ** 2, dtype=numpy.complex64)
        data.shape = [3] * 2
        widget = self.create_widget()
        widget.setData(data)
        availableModes = set([v.modeId() for v in widget.currentAvailableViews()])
        self.assertEqual(DataViews.RAW_MODE, widget.displayMode())
        self.assertIn(DataViews.IMAGE_MODE, availableModes)

    def test_plot_3d_data(self):
        data = numpy.arange(3 ** 3)
        data.shape = [3] * 3
        widget = self.create_widget()
        widget.setData(data)
        availableModes = set([v.modeId() for v in widget.currentAvailableViews()])
        try:
            import silx.gui.plot3d  # noqa
            self.assertIn(DataViews.PLOT3D_MODE, availableModes)
        except ImportError:
            self.assertIn(DataViews.STACK_MODE, availableModes)
        self.assertEqual(DataViews.RAW_MODE, widget.displayMode())

    def test_array_1d_data(self):
        data = numpy.array(["aaa"] * (3 ** 1))
        data.shape = [3] * 1
        widget = self.create_widget()
        widget.setData(data)
        self.assertEqual(DataViews.RAW_MODE, widget.displayedView().modeId())

    def test_array_2d_data(self):
        data = numpy.array(["aaa"] * (3 ** 2))
        data.shape = [3] * 2
        widget = self.create_widget()
        widget.setData(data)
        self.assertEqual(DataViews.RAW_MODE, widget.displayedView().modeId())

    def test_array_4d_data(self):
        data = numpy.array(["aaa"] * (3 ** 4))
        data.shape = [3] * 4
        widget = self.create_widget()
        widget.setData(data)
        self.assertEqual(DataViews.RAW_MODE, widget.displayedView().modeId())

    def test_record_4d_data(self):
        data = numpy.zeros(3 ** 4, dtype='3int8, float32, (2,3)float64')
        data.shape = [3] * 4
        widget = self.create_widget()
        widget.setData(data)
        self.assertEqual(DataViews.RAW_MODE, widget.displayedView().modeId())

    def test_3d_h5_dataset(self):
        with self.h5_temporary_file() as h5file:
            dataset = h5file["data"]
            widget = self.create_widget()
            widget.setData(dataset)

    def test_data_event(self):
        listener = SignalListener()
        widget = self.create_widget()
        widget.dataChanged.connect(listener)
        widget.setData(10)
        widget.setData(None)
        self.assertEqual(listener.callCount(), 2)

    def test_display_mode_event(self):
        listener = SignalListener()
        widget = self.create_widget()
        widget.displayedViewChanged.connect(listener)
        widget.setData(10)
        widget.setData(None)
        modes = [v.modeId() for v in listener.arguments(argumentIndex=0)]
        self.assertEqual(modes, [DataViews.RAW_MODE, DataViews.EMPTY_MODE])
        listener.clear()

    def test_change_display_mode(self):
        listener = SignalListener()
        data = numpy.arange(10 ** 4)
        data.shape = [10] * 4
        widget = self.create_widget()
        widget.selectionChanged.connect(listener)
        widget.setData(data)

        widget.setDisplayMode(DataViews.PLOT1D_MODE)
        self.assertEqual(widget.displayedView().modeId(), DataViews.PLOT1D_MODE)
        self.qWait(200)
        assert listener.arguments() == [((0, 0, 0, slice(None)), None)]
        listener.clear()

        widget.setDisplayMode(DataViews.IMAGE_MODE)
        self.assertEqual(widget.displayedView().modeId(), DataViews.IMAGE_MODE)
        self.qWait(200)
        assert listener.arguments() == [((0, 0, slice(None), slice(None)), None)]
        listener.clear()

        widget.setDisplayMode(DataViews.RAW_MODE)
        self.assertEqual(widget.displayedView().modeId(), DataViews.RAW_MODE)
        self.qWait(200)
        # Changing from 2D to 2D view: Selection didn't changed
        assert listener.callCount() == 0

        widget.setDisplayMode(DataViews.EMPTY_MODE)
        self.assertEqual(widget.displayedView().modeId(), DataViews.EMPTY_MODE)
        self.qWait(200)
        assert listener.arguments() == [(None, None)]
        listener.clear()

    def test_create_default_views(self):
        widget = self.create_widget()
        views = widget.createDefaultViews()
        self.assertTrue(len(views) > 0)

    def test_add_view(self):
        widget = self.create_widget()
        view = _DataViewMock(widget)
        widget.addView(view)
        self.assertTrue(view in widget.availableViews())
        self.assertTrue(view in widget.currentAvailableViews())

    def test_remove_view(self):
        widget = self.create_widget()
        widget.setData("foobar")
        view = widget.currentAvailableViews()[0]
        widget.removeView(view)
        self.assertTrue(view not in widget.availableViews())
        self.assertTrue(view not in widget.currentAvailableViews())

    def test_replace_view(self):
        widget = self.create_widget()
        view = _DataViewMock(widget)
        widget.replaceView(DataViews.RAW_MODE,
                           view)
        self.assertIsNone(widget.getViewFromModeId(DataViews.RAW_MODE))
        self.assertTrue(view in widget.availableViews())
        self.assertTrue(view in widget.currentAvailableViews())

    def test_replace_view_in_composite(self):
        # replace a view that is a child of a composite view
        widget = self.create_widget()
        view = _DataViewMock(widget)
        replaced = widget.replaceView(DataViews.NXDATA_INVALID_MODE,
                                      view)
        self.assertTrue(replaced)
        nxdata_view = widget.getViewFromModeId(DataViews.NXDATA_MODE)
        self.assertNotIn(DataViews.NXDATA_INVALID_MODE,
                         [v.modeId() for v in nxdata_view.getViews()])
        self.assertTrue(view in nxdata_view.getViews())


class TestDataViewer(_TestAbstractDataViewer):
    __test__ = True  # because _TestAbstractDataViewer is ignored
    def create_widget(self):
        return DataViewer()


class TestDataViewerFrame(_TestAbstractDataViewer):
    __test__ = True  # because _TestAbstractDataViewer is ignored
    def create_widget(self):
        return DataViewerFrame()


class TestDataView(TestCaseQt):

    def createComplexData(self):
        line = [1, 2j, 3 + 3j, 4]
        image = [line, line, line, line]
        cube = [image, image, image, image]
        data = numpy.array(cube, dtype=numpy.complex64)
        return data

    def createDataViewWithData(self, dataViewClass, data):
        viewer = dataViewClass(None)
        widget = viewer.getWidget()
        viewer.setData(data)
        return widget

    def testCurveWithComplex(self):
        data = self.createComplexData()
        dataViewClass = DataViews._Plot1dView
        widget = self.createDataViewWithData(dataViewClass, data[0, 0])
        self.qWaitForWindowExposed(widget)

    def testImageWithComplex(self):
        data = self.createComplexData()
        dataViewClass = DataViews._Plot2dView
        widget = self.createDataViewWithData(dataViewClass, data[0])
        self.qWaitForWindowExposed(widget)

    @pytest.mark.usefixtures("use_opengl")
    def testCubeWithComplex(self):
        try:
            import silx.gui.plot3d  # noqa
        except ImportError:
            self.skipTest("OpenGL not available")
        data = self.createComplexData()
        dataViewClass = DataViews._Plot3dView
        widget = self.createDataViewWithData(dataViewClass, data)
        self.qWaitForWindowExposed(widget)

    def testImageStackWithComplex(self):
        data = self.createComplexData()
        dataViewClass = DataViews._StackView
        widget = self.createDataViewWithData(dataViewClass, data)
        self.qWaitForWindowExposed(widget)



# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Sphinx extension creating a link to an archive.

This Sphinx extension adds the possibility to create a zip archive and
include it in the generated HTML documentation.

This extension provides an *archive* directive.

Usage:

.. archive:: <relative_path/to/directory/to/compress>
   :filename: <name_of_the_archive_file, default: last_folder_of_directory.zip>
   :filter: <Space-separated list of included file patterns, default *.*>
   :basedir: <Name of the base directory in the archive, default: filename>

Examples:

To create a example.zip archive of the ../../example/ folder:

.. archive:: ../../examples/

To get more control on the name of the archive and its content:

.. archive:: ../../examples/
   :filename: myproject_examples.zip
   :filter: *.py *.png
   :basedir: myproject_examples

WARNING: The content of this directory is not checked for outdated documents.
"""
__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "01/09/2017"


import fnmatch
import logging
import os
import os.path
import posixpath
import shutil
import tempfile

from docutils.parsers.rst import directives, Directive
import docutils.nodes


_logger = logging.getLogger(__name__)


# docutils directive

class ArchiveDirective(Directive):
    """Add a link to download an archive

    This directive add an :class:`archive` node to the document tree.
    """

    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {
        'filename': directives.unchanged,
        'filter': directives.unchanged,
        'basedir': directives.unchanged,
    }

    def run(self):
        node = archive('')

        # Get rst source file containing this directive
        source_file = self.state_machine.get_source_and_line()[0]
        if source_file is None:
            raise RuntimeError('Cannot get rst source file path')

        # Build input path from rst source file and directive argument
        input_path = self.arguments[0]
        if not input_path.startswith('/'):  # Argument is a relative path
            input_path = os.path.abspath(
                os.path.join(os.path.dirname(source_file), input_path))
        node['input_path'] = input_path

        default_basedir = os.path.basename(input_path)
        node['basedir'] = self.options.get('basedir', default_basedir)
        node['filename'] = self.options.get('filename',
                                            '.'.join((default_basedir, 'zip')))

        node['filter'] = self.options.get('filter', '*.*')

        return [node]


# archive doctuils node

class archive(docutils.nodes.General, docutils.nodes.Element, docutils.nodes.Inline):
    """archive node created by :class:`ArchiveDirective`"""
    pass


def visit_archive_html(self, node):
    """Node visitor to translate :class:`archive` nodes to HTML.

    :param self: Sphinx HTML Writter
    :param node: The :class:`archive` node to translate to HTML
    :raise: SkipNode as depart is not implemented
    """
    filename = node['filename']
    input_path = node['input_path']

    # Create a temporary folder to create archive content
    tmp_dir = tempfile.mkdtemp()

    # Copy selected content to temporary folder
    base_dir = os.path.join(tmp_dir, node['basedir'])

    def ignore(src, names):
        patterns = node['filter'].split()
        ignored_names = []
        for name in names:
            for pattern in patterns:
                if fnmatch.fnmatch(name, pattern):
                    break
            else:
                ignored_names.append(name)
        return ignored_names

    shutil.copytree(input_path, base_dir, ignore=ignore)

    # Compress temporary folder to zip
    output_filename = os.path.join(
        self.builder.outdir, '_downloads', filename)
    root, ext = os.path.splitext(output_filename)
    assert ext == '.zip'
    shutil.make_archive(root, 'zip', tmp_dir, node['basedir'])

    # Clean-up temporary folder
    shutil.rmtree(tmp_dir)

    # Generate HTML
    relative_path = posixpath.join(self.builder.dlpath, filename)
    self.body.append('<a href="%s">%s</a>' % (relative_path, filename))
    raise docutils.nodes.SkipNode


def visit_skip(self, node):
    """No-op node visitor"""
    raise docutils.nodes.SkipNode


# Extension setup

def setup(app):
    """Sphinx extension registration"""
    app.add_node(archive,
                 html=(visit_archive_html, None),
                 latex=(visit_skip, None))

    app.add_directive('archive', ArchiveDirective)

    return {'version': '0.1'}



# /*##########################################################################
#
# Copyright (c) 2015-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""This module provides functions making masks on an image.

- :func:`circle_fill` function generates coordinates of a circle in an image.
- :func:`draw_line` function generates coordinates of a line in an image.
- :func:`polygon_fill_mask` function generates a mask from a set of points
  defining a polygon.

The :class:`Polygon` class provides checking if a point is inside a polygon.

The whole module uses the (row, col) (i.e., (y, x))) convention
for 2D coordinates.
"""

__authors__ = ["Jérôme Kieffer", "T. Vincent"]
__license__ = "MIT"
__date__ = "15/02/2019"
__status__ = "dev"


cimport cython
import numpy
from libc.math cimport ceil, fabs


cdef class Polygon(object):
    """Define a polygon that provides inside check and mask generation.

    :param vertices: corners of the polygon
    :type vertices: Nx2 array of floats of (row, col)
    """

    cdef float[:,:] vertices
    cdef int nvert

    def __init__(self, vertices):
        self.vertices = numpy.ascontiguousarray(vertices, dtype=numpy.float32)
        if self.vertices.ndim != 2 or self.vertices.shape[1] != 2:
             raise ValueError("A list of 2d vertices is expected (n,2 dimensional array)")
        self.nvert = self.vertices.shape[0]

    def is_inside(self, row, col):
        """Check if (row, col) is inside or outside the polygon

        :param float row:
        :param float col:
        :return: True if position is inside polygon, False otherwise
        """
        return self.c_is_inside(row, col)

    @cython.cdivision(True)
    @cython.wraparound(False)
    @cython.boundscheck(False)
    cdef bint c_is_inside(self, float row, float col) nogil:
        """Check if (row, col) is inside or outside the polygon

        Pure C_Cython class implementation.

        :param float row:
        :param float col:
        :return: True if position is inside polygon, False otherwise
        """
        cdef int index, is_inside
        cdef float pt1x, pt1y, pt2x, pt2y, xinters
        is_inside = 0

        pt1x = self.vertices[self.nvert-1, 1]
        pt1y = self.vertices[self.nvert-1, 0]
        for index in range(self.nvert):
            pt2x = self.vertices[index, 1]
            pt2y = self.vertices[index, 0]

            if (((pt1y <= row and row < pt2y) or
                    (pt2y <= row and row < pt1y)) and
                    # Extra (optional) condition to avoid some computation
                    (col <= pt1x or col <= pt2x)):
                xinters = (row - pt1y) * (pt2x - pt1x) / (pt2y - pt1y) + pt1x
                is_inside ^= col < xinters
            pt1x, pt1y = pt2x, pt2y
        return is_inside

    @cython.cdivision(True)
    @cython.wraparound(False)
    @cython.boundscheck(False)
    def make_mask(self, int height, int width):
        """Create a mask array representing the filled polygon

        :param int height: Height of the mask array
        :param int width: Width of the mask array
        :return: 2D array (height, width)
        """
        cdef unsigned char[:, :] mask = numpy.zeros((height, width),
                                                    dtype=numpy.uint8)
        cdef int row_min, row_max, col_min, col_max  # mask subpart to update
        cdef int row, col, index  # Loop indixes
        cdef float pt1x, pt1y, pt2x, pt2y  # segment end points
        cdef int xinters, is_inside, current

        row_min = max(int(min(self.vertices[:, 0])), 0)
        row_max = min(int(max(self.vertices[:, 0])) + 1, height)

        # Can be replaced by prange(row_min, row_max, nogil=True)
        with nogil:
            for row in range(row_min, row_max):
                # For each line of the image, mark intersection of all segments
                # in the line and then run a xor scan to fill inner parts
                # Adapted from http://alienryderflex.com/polygon_fill/
                pt1x = self.vertices[self.nvert-1, 1]
                pt1y = self.vertices[self.nvert-1, 0]
                col_min = width - 1
                col_max = 0
                is_inside = 0  # Init with whether first col is inside or not

                for index in range(self.nvert):
                    pt2x = self.vertices[index, 1]
                    pt2y = self.vertices[index, 0]

                    if ((pt1y <= row and row < pt2y) or
                            (pt2y <= row and row < pt1y)):
                        # Intersection casted to int so that ]x, x+1] => x
                        xinters = (<int>ceil(pt1x + (row - pt1y) *
                                   (pt2x - pt1x) / (pt2y - pt1y))) - 1

                        # Update column range to patch
                        if xinters < col_min:
                            col_min = xinters
                        if xinters > col_max:
                            col_max = xinters

                        if xinters < 0:
                            # Add an intersection to init value of xor scan
                            is_inside ^= 1
                        elif xinters < width:
                            # Mark intersection in mask
                            mask[row, xinters] ^= 1
                        # else: do not consider intersection on the right

                    pt1x, pt1y = pt2x, pt2y

                if col_min < col_max:
                    # Clip column range to mask
                    if col_min < 0:
                        col_min = 0
                    if col_max > width - 1:
                        col_max = width - 1

                    # xor exclusive scan
                    for col in range(col_min, col_max + 1):
                        current = mask[row, col]
                        mask[row, col] = is_inside
                        is_inside = current ^ is_inside

        # Ensures the result is exported as numpy array and not memory view.
        return numpy.asarray(mask)


def polygon_fill_mask(vertices, shape):
    """Return a mask of boolean, True for pixels inside a polygon.

    :param vertices: Strip of segments end points (row, column) or (y, x)
    :type vertices: numpy.ndarray like container of dimension Nx2
    :param shape: size of the mask as (height, width)
    :type shape: 2-tuple of int
    :return: Mask corresponding to the polygon
    :rtype: numpy.ndarray of dimension shape
    """
    return Polygon(vertices).make_mask(shape[0], shape[1])


@cython.wraparound(False)
@cython.boundscheck(False)
def draw_line(int row0, int col0, int row1, int col1, int width=1):
    """Line includes both end points.
    Width is handled by drawing parallel lines, so junctions of lines belonging
    to different octant with width > 1 will not look nice.

    Using Bresenham line algorithm:
    Bresenham, J. E.
    Algorithm for computer control of a digital plotter.
    IBM Systems Journal. Vol 4 No 1. 1965. pp 25-30

    :param int row0: Start point row
    :param int col0: Start point col
    :param int row1: End point row
    :param int col1: End point col
    :param int width: Thickness of the line in pixels (default 1)
                      Width must be at least 1.
    :return: Array coordinates of points inside the line (might be negative)
    :rtype: 2-tuple of numpy.ndarray (rows, cols)
    """
    cdef int drow, dcol, invert_coords
    cdef int db, da, delta, b, a, step_a, step_b
    cdef int index, offset  # Loop indices

    # Store coordinates of points of the line
    cdef int[:, :] b_coords
    cdef int[:, :] a_coords

    dcol = abs(col1 - col0)
    drow = abs(row1 - row0)
    invert_coords = dcol < drow

    if dcol == 0 and drow == 0:
        return (numpy.array((row0,), dtype=numpy.int32),
                numpy.array((col0,), dtype=numpy.int32))

    if width < 1:
        width = 1

    # Set a and b according to segment octant
    if not invert_coords:
        da = dcol
        db = drow
        step_a = 1 if col1 > col0 else -1
        step_b = 1 if row1 > row0 else -1
        a = col0
        b = row0

    else:
        da = drow
        db = dcol
        step_a = 1 if row1 > row0 else -1
        step_b = 1 if col1 > col0 else -1
        a = row0
        b = col0

    b_coords = numpy.empty((da + 1, width), dtype=numpy.int32)
    a_coords = numpy.empty((da + 1, width), dtype=numpy.int32)

    with nogil:
        b -= (width - 1) // 2
        delta = 2 * db - da
        for index in range(da + 1):
            for offset in range(width):
                b_coords[index, offset] = b + offset
                a_coords[index, offset] = a

            if delta >= 0:  # M2: Move by step_a + step_b
                b += step_b
                delta -= 2 * da
            # else M1: Move by step_a

            a += step_a
            delta += 2 * db

    if not invert_coords:
        return (numpy.asarray(b_coords).reshape(-1),
                numpy.asarray(a_coords).reshape(-1))
    else:
        return (numpy.asarray(a_coords).reshape(-1),
                numpy.asarray(b_coords).reshape(-1))


def circle_fill(int crow, int ccol, float radius):
    """Generates coordinate of image points lying in a disk.

    :param int crow: Row of the center of the disk
    :param int ccol: Column of the center of the disk
    :param float radius: Radius of the disk
    :return: Array coordinates of points inside the disk (might be negative)
    :rtype: 2-tuple of numpy.ndarray (rows, cols)
    """
    cdef int i_radius, len_coords

    radius = fabs(radius)
    i_radius = <int>radius

    coords = numpy.arange(-i_radius, ceil(radius) + 1,
                          dtype=numpy.float32) ** 2
    len_coords = len(coords)
    # rows, cols = where(row**2 + col**2 < radius**2)
    rows, cols = numpy.where(coords.reshape(1, len_coords) +
                             coords.reshape(len_coords, 1) < radius ** 2)
    return rows + crow - i_radius, cols + ccol - i_radius


def ellipse_fill(int crow, int ccol, float radius_r, float radius_c):
    """Generates coordinate of image points lying in a ellipse.

    :param int crow: Row of the center of the ellipse
    :param int ccol: Column of the center of the ellipse
    :param float radius_r: Radius of the ellipse in the row
    :param float radius_c: Radius of the ellipse in the column
    :return: Array coordinates of points inside the ellipse (might be negative)
    :rtype: 2-tuple of numpy.ndarray (rows, cols)
    """
    cdef int i_radius_r
    cdef int i_radius_c

    i_radius_r = <int>fabs(radius_r)
    i_radius_c = <int>fabs(radius_c)

    x_coords = numpy.arange(-i_radius_r, ceil(radius_r) + 1, dtype=numpy.float32).reshape(-1, 1)
    y_coords = numpy.arange(-i_radius_c, ceil(radius_c) + 1, dtype=numpy.float32).reshape(1, -1)

    # rows, cols = where(x**2 + col**2 < 1)
    rows, cols = numpy.where(x_coords**2 / radius_r**2 + y_coords**2 / radius_c**2 < 1.0)
    return rows + crow - i_radius_r, cols + ccol - i_radius_c



# /*##########################################################################
#
# Copyright (c) 2017-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This script shows the features of a :mod:`~silx.gui.dialog.ColormapDialog`.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "14/06/2018"

import functools
import numpy

try:
    import scipy
except ImportError:
    scipy = None

from silx.gui import qt
from silx.gui.dialog.ColormapDialog import ColormapDialog
from silx.gui.colors import Colormap
from silx.gui.plot.ColorBar import ColorBarWidget


class ColormapDialogExample(qt.QMainWindow):
    """PlotWidget with an ad hoc toolbar and a colorbar"""

    def __init__(self, parent=None):
        super(ColormapDialogExample, self).__init__(parent)
        self.setWindowTitle("Colormap dialog example")

        self.colormap1 = Colormap("viridis")
        self.colormap2 = Colormap("gray")

        self.colorBar = ColorBarWidget(self)

        self.colorDialogs = []

        options = qt.QWidget(self)
        options.setLayout(qt.QVBoxLayout())
        self.createOptions(options.layout())

        mainWidget = qt.QWidget(self)
        mainWidget.setLayout(qt.QHBoxLayout())
        mainWidget.layout().addWidget(options)
        mainWidget.layout().addWidget(self.colorBar)
        self.mainWidget = mainWidget

        self.setCentralWidget(mainWidget)
        self.createColorDialog()

    def createOptions(self, layout):
        button = qt.QPushButton("Create a new dialog")
        button.clicked.connect(self.createColorDialog)
        layout.addWidget(button)

        layout.addSpacing(10)

        button = qt.QPushButton("Set editable")
        button.clicked.connect(self.setEditable)
        layout.addWidget(button)
        button = qt.QPushButton("Set non-editable")
        button.clicked.connect(self.setNonEditable)
        layout.addWidget(button)

        layout.addSpacing(10)

        button = qt.QPushButton("Set no colormap")
        button.clicked.connect(self.setNoColormap)
        layout.addWidget(button)
        button = qt.QPushButton("Set colormap 1")
        button.clicked.connect(self.setColormap1)
        layout.addWidget(button)
        button = qt.QPushButton("Set colormap 2")
        button.clicked.connect(self.setColormap2)
        layout.addWidget(button)
        button = qt.QPushButton("Create new colormap")
        button.clicked.connect(self.setNewColormap)
        layout.addWidget(button)

        layout.addSpacing(10)

        button = qt.QPushButton("No histogram")
        button.clicked.connect(self.setNoHistogram)
        layout.addWidget(button)
        button = qt.QPushButton("Positive histogram")
        button.clicked.connect(self.setPositiveHistogram)
        layout.addWidget(button)
        button = qt.QPushButton("Neg-pos histogram")
        button.clicked.connect(self.setNegPosHistogram)
        layout.addWidget(button)
        button = qt.QPushButton("Negative histogram")
        button.clicked.connect(self.setNegativeHistogram)
        layout.addWidget(button)

        layout.addSpacing(10)

        button = qt.QPushButton("No range")
        button.clicked.connect(self.setNoRange)
        layout.addWidget(button)
        button = qt.QPushButton("Positive range")
        button.clicked.connect(self.setPositiveRange)
        layout.addWidget(button)
        button = qt.QPushButton("Neg-pos range")
        button.clicked.connect(self.setNegPosRange)
        layout.addWidget(button)
        button = qt.QPushButton("Negative range")
        button.clicked.connect(self.setNegativeRange)
        layout.addWidget(button)

        layout.addSpacing(10)

        button = qt.QPushButton("No data")
        button.clicked.connect(self.setNoData)
        layout.addWidget(button)
        button = qt.QPushButton("Zero to positive")
        button.clicked.connect(self.setSheppLoganPhantom)
        layout.addWidget(button)
        button = qt.QPushButton("Negative to positive")
        button.clicked.connect(self.setDataFromNegToPos)
        layout.addWidget(button)
        button = qt.QPushButton("With non finite values")
        button.clicked.connect(self.setDataWithNonFinite)
        layout.addWidget(button)

        layout.addStretch()

    def createColorDialog(self):
        newDialog = ColormapDialog(self)
        newDialog.finished.connect(functools.partial(self.removeColorDialog, newDialog))
        self.colorDialogs.append(newDialog)
        self.mainWidget.layout().addWidget(newDialog)

    def removeColorDialog(self, dialog, result):
        self.colorDialogs.remove(dialog)

    def setNoColormap(self):
        self.colorBar.setColormap(None)
        for dialog in self.colorDialogs:
            dialog.setColormap(None)

    def setColormap1(self):
        self.colorBar.setColormap(self.colormap1)
        for dialog in self.colorDialogs:
            dialog.setColormap(self.colormap1)

    def setColormap2(self):
        self.colorBar.setColormap(self.colormap2)
        for dialog in self.colorDialogs:
            dialog.setColormap(self.colormap2)

    def setEditable(self):
        for dialog in self.colorDialogs:
            colormap = dialog.getColormap()
            if colormap is not None:
                colormap.setEditable(True)

    def setNonEditable(self):
        for dialog in self.colorDialogs:
            colormap = dialog.getColormap()
            if colormap is not None:
                colormap.setEditable(False)

    def setNewColormap(self):
        self.colormap = Colormap("inferno")
        self.colorBar.setColormap(self.colormap)
        for dialog in self.colorDialogs:
            dialog.setColormap(self.colormap)

    def setNoHistogram(self):
        for dialog in self.colorDialogs:
            dialog.setHistogram()

    def setPositiveHistogram(self):
        histo = [5, 10, 50, 10, 5]
        pos = 1
        edges = list(range(pos, pos + len(histo)))
        for dialog in self.colorDialogs:
            dialog.setHistogram(histo, edges)

    def setNegPosHistogram(self):
        histo = [5, 10, 50, 10, 5]
        pos = -2
        edges = list(range(pos, pos + len(histo)))
        for dialog in self.colorDialogs:
            dialog.setHistogram(histo, edges)

    def setNegativeHistogram(self):
        histo = [5, 10, 50, 10, 5]
        pos = -30
        edges = list(range(pos, pos + len(histo)))
        for dialog in self.colorDialogs:
            dialog.setHistogram(histo, edges)

    def setNoRange(self):
        for dialog in self.colorDialogs:
            dialog.setDataRange()

    def setPositiveRange(self):
        for dialog in self.colorDialogs:
            dialog.setDataRange(1, 1, 10)

    def setNegPosRange(self):
        for dialog in self.colorDialogs:
            dialog.setDataRange(-10, 1, 10)

    def setNegativeRange(self):
        for dialog in self.colorDialogs:
            dialog.setDataRange(-10, float("nan"), -1)

    def setNoData(self):
        for dialog in self.colorDialogs:
            dialog.setData(None)

    def setSheppLoganPhantom(self):
        from silx.image import phantomgenerator
        data = phantomgenerator.PhantomGenerator.get2DPhantomSheppLogan(256)
        data = data * 1000
        if scipy is not None:
            from scipy import ndimage
            data = ndimage.gaussian_filter(data, sigma=20)
        data = numpy.random.poisson(data)
        self.data = data
        for dialog in self.colorDialogs:
            dialog.setData(self.data)

    def setDataFromNegToPos(self):
        data = numpy.ones((50,50))
        data = numpy.random.poisson(data)
        self.data = data - 0.5
        for dialog in self.colorDialogs:
            dialog.setData(self.data)

    def setDataWithNonFinite(self):
        from silx.image import phantomgenerator
        data = phantomgenerator.PhantomGenerator.get2DPhantomSheppLogan(256)
        data = data * 1000
        if scipy is not None:
            from scipy import ndimage
            data = ndimage.gaussian_filter(data, sigma=20)
        data = numpy.random.poisson(data).astype(numpy.float32)
        data[10] = float("nan")
        data[50] = float("+inf")
        data[100] = float("-inf")
        self.data = data
        for dialog in self.colorDialogs:
            dialog.setData(self.data)


def main():
    app = qt.QApplication([])

    # Create the ad hoc plot widget and change its default colormap
    example = ColormapDialogExample()
    example.show()

    app.exec()


if __name__ == '__main__':
    main()



# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
This module contains utilitaries used by other dialog modules.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "17/01/2019"

import fabio
import silx.io
from silx.gui import qt


class Codec(object):

    def __init__(self, any_fabio=False, any_silx=False, fabio_codec=None, auto=False):
        self.__any_fabio = any_fabio
        self.__any_silx = any_silx
        self.fabio_codec = fabio_codec
        self.__auto = auto

    def is_autodetect(self):
        return self.__auto

    def is_fabio_codec(self):
        return self.__any_fabio or self.fabio_codec is not None

    def is_silx_codec(self):
        return self.__any_silx


class FileTypeComboBox(qt.QComboBox):
    """
    A combobox providing all image file formats supported by fabio and silx.

    It provides access for each fabio codecs individually.
    """

    EXTENSIONS_ROLE = qt.Qt.UserRole + 1

    CODEC_ROLE = qt.Qt.UserRole + 2

    INDENTATION = u"\u2022 "

    def __init__(self, parent=None):
        qt.QComboBox.__init__(self, parent)
        self.__fabioUrlSupported = True
        self.__initItems()

    def setFabioUrlSupproted(self, isSupported):
        if self.__fabioUrlSupported == isSupported:
            return
        self.__fabioUrlSupported = isSupported
        self.__initItems()

    def __initItems(self):
        self.clear()
        if self.__fabioUrlSupported:
            self.__insertFabioFormats()
        self.__insertSilxFormats()
        self.__insertAllSupported()
        self.__insertAnyFiles()

    def __insertAnyFiles(self):
        index = self.count()
        self.addItem("All files (*)")
        self.setItemData(index, ["*"], role=self.EXTENSIONS_ROLE)
        self.setItemData(index, Codec(auto=True), role=self.CODEC_ROLE)

    def __insertAllSupported(self):
        allExtensions = set([])
        for index in range(self.count()):
            ext = self.itemExtensions(index)
            allExtensions.update(ext)
        allExtensions = allExtensions - set("*")
        list(sorted(list(allExtensions)))
        index = 0
        self.insertItem(index, "All supported files")
        self.setItemData(index, allExtensions, role=self.EXTENSIONS_ROLE)
        self.setItemData(index, Codec(auto=True), role=self.CODEC_ROLE)

    def __insertSilxFormats(self):
        formats = silx.io.supported_extensions()

        extensions = []
        allExtensions = set([])

        for description, ext in formats.items():
            allExtensions.update(ext)
            if ext == []:
                ext = ["*"]
            extensions.append((description, ext, "silx"))
        extensions = list(sorted(extensions))

        allExtensions = list(sorted(list(allExtensions)))
        index = self.count()
        self.addItem("All supported files, using Silx")
        self.setItemData(index, allExtensions, role=self.EXTENSIONS_ROLE)
        self.setItemData(index, Codec(any_silx=True), role=self.CODEC_ROLE)

        for e in extensions:
            index = self.count()
            if len(e[1]) < 10:
                self.addItem("%s%s (%s)" % (self.INDENTATION, e[0], " ".join(e[1])))
            else:
                self.addItem("%s%s" % (self.INDENTATION, e[0]))
            codec = Codec(any_silx=True)
            self.setItemData(index, e[1], role=self.EXTENSIONS_ROLE)
            self.setItemData(index, codec, role=self.CODEC_ROLE)

    def __insertFabioFormats(self):
        formats = fabio.fabioformats.get_classes(reader=True)

        from fabio import fabioutils
        if hasattr(fabioutils, "COMPRESSED_EXTENSIONS"):
            compressedExtensions = fabioutils.COMPRESSED_EXTENSIONS
        else:
            # Support for fabio < 0.9
            compressedExtensions = set(["gz", "bz2"])

        extensions = []
        allExtensions = set([])

        def extensionsIterator(reader):
            for extension in reader.DEFAULT_EXTENSIONS:
                yield "*.%s" % extension
            for compressedExtension in compressedExtensions:
                for extension in reader.DEFAULT_EXTENSIONS:
                    yield "*.%s.%s" % (extension, compressedExtension)

        for reader in formats:
            if not hasattr(reader, "DESCRIPTION"):
                continue
            if not hasattr(reader, "DEFAULT_EXTENSIONS"):
                continue

            displayext = reader.DEFAULT_EXTENSIONS
            displayext = ["*.%s" % e for e in displayext]
            ext = list(extensionsIterator(reader))
            allExtensions.update(ext)
            if ext == []:
                ext = ["*"]
            extensions.append((reader.DESCRIPTION, displayext, ext, reader.codec_name()))
        extensions = list(sorted(extensions))

        allExtensions = list(sorted(list(allExtensions)))
        index = self.count()
        self.addItem("All supported files, using Fabio")
        self.setItemData(index, allExtensions, role=self.EXTENSIONS_ROLE)
        self.setItemData(index, Codec(any_fabio=True), role=self.CODEC_ROLE)

        for e in extensions:
            description, displayExt, allExt, _codecName = e
            index = self.count()
            if len(e[1]) < 10:
                self.addItem("%s%s (%s)" % (self.INDENTATION, description, " ".join(displayExt)))
            else:
                self.addItem("%s%s" % (self.INDENTATION, description))
            codec = Codec(fabio_codec=_codecName)
            self.setItemData(index, allExt, role=self.EXTENSIONS_ROLE)
            self.setItemData(index, codec, role=self.CODEC_ROLE)

    def itemExtensions(self, index):
        """Returns the extensions associated to an index."""
        result = self.itemData(index, self.EXTENSIONS_ROLE)
        if result is None:
            result = None
        return result

    def currentExtensions(self):
        """Returns the current selected extensions."""
        index = self.currentIndex()
        return self.itemExtensions(index)

    def indexFromCodec(self, codecName):
        for i in range(self.count()):
            codec = self.itemCodec(i)
            if codecName == "auto":
                if codec.is_autodetect():
                    return i
            elif codecName == "silx":
                if codec.is_silx_codec():
                    return i
            elif codecName == "fabio":
                if codec.is_fabio_codec() and codec.fabio_codec is None:
                    return i
            elif codecName == codec.fabio_codec:
                return i
        return -1

    def itemCodec(self, index):
        """Returns the codec associated to an index."""
        result = self.itemData(index, self.CODEC_ROLE)
        if result is None:
            result = None
        return result

    def currentCodec(self):
        """Returns the current selected codec. None if nothing selected
        or if the item is not a codec"""
        index = self.currentIndex()
        return self.itemCodec(index)



# /*##########################################################################
#
# Copyright (c) 2018-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""This module provides helper functions for applying colormaps to datasets"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "25/08/2021"


import collections
import warnings
import numpy

from ..resources import resource_filename as _resource_filename
from .combo import min_max as _min_max
from . import _colormap
from ._colormap import cmap  # noqa


__all__ = ["apply_colormap", "cmap"]


_LUT_DESCRIPTION = collections.namedtuple("_LUT_DESCRIPTION", ["source", "cursor_color"])
"""Description of a LUT for internal purpose."""


_AVAILABLE_LUTS = collections.OrderedDict([
    ('gray', _LUT_DESCRIPTION('builtin', '#ff66ff')),
    ('reversed gray', _LUT_DESCRIPTION('builtin', '#ff66ff')),
    ('red', _LUT_DESCRIPTION('builtin', '#00ff00')),
    ('green', _LUT_DESCRIPTION('builtin', '#ff66ff')),
    ('blue', _LUT_DESCRIPTION('builtin', '#ffff00')),
    ('viridis', _LUT_DESCRIPTION('resource', '#ff66ff')),
    ('cividis', _LUT_DESCRIPTION('resource', '#ff66ff')),
    ('magma', _LUT_DESCRIPTION('resource', '#00ff00')),
    ('inferno', _LUT_DESCRIPTION('resource', '#00ff00')),
    ('plasma', _LUT_DESCRIPTION('resource', '#00ff00')),
    ('temperature', _LUT_DESCRIPTION('builtin', '#ff66ff')),
])
"""Description for internal porpose of all the default LUT provided by the library."""


# Colormap loader

_COLORMAP_CACHE = {}
"""Cache already used colormaps as name: color LUT"""


def array_to_rgba8888(colors):
    """Convert colors from a numpy array using float (0..1) int or uint
    (0..255) to uint8 RGBA.

    :param numpy.ndarray colors: Array of float int or uint  colors to convert
    :return: colors as uint8
    :rtype: numpy.ndarray
    """
    assert len(colors.shape) == 2
    assert colors.shape[1] in (3, 4)

    if colors.dtype == numpy.uint8:
        pass
    elif colors.dtype.kind == 'f':
        # Each bin is [N, N+1[ except the last one: [255, 256]
        colors = numpy.clip(colors.astype(numpy.float64) * 256, 0., 255.)
        colors = colors.astype(numpy.uint8)
    elif colors.dtype.kind in 'iu':
        colors = numpy.clip(colors, 0, 255)
        colors = colors.astype(numpy.uint8)

    if colors.shape[1] == 3:
        tmp = numpy.empty((len(colors), 4), dtype=numpy.uint8)
        tmp[:, 0:3] = colors
        tmp[:, 3] = 255
        colors = tmp

    return colors


def _create_colormap_lut(name):
    """Returns the color LUT corresponding to a colormap name

    :param str name: Name of the colormap to load
    :returns: Corresponding table of colors
    :rtype: numpy.ndarray
    :raise ValueError: If no colormap corresponds to name
    """
    description = _AVAILABLE_LUTS.get(name)
    if description is not None:
        if description.source == "builtin":
            # Build colormap LUT
            lut = numpy.zeros((256, 4), dtype=numpy.uint8)
            lut[:, 3] = 255

            if name == 'gray':
                lut[:, :3] = numpy.arange(256, dtype=numpy.uint8).reshape(-1, 1)
            elif name == 'reversed gray':
                lut[:, :3] = numpy.arange(255, -1, -1, dtype=numpy.uint8).reshape(-1, 1)
            elif name == 'red':
                lut[:, 0] = numpy.arange(256, dtype=numpy.uint8)
            elif name == 'green':
                lut[:, 1] = numpy.arange(256, dtype=numpy.uint8)
            elif name == 'blue':
                lut[:, 2] = numpy.arange(256, dtype=numpy.uint8)
            elif name == 'temperature':
                # Red
                lut[128:192, 0] = numpy.arange(2, 255, 4, dtype=numpy.uint8)
                lut[192:, 0] = 255
                # Green
                lut[:64, 1] = numpy.arange(0, 255, 4, dtype=numpy.uint8)
                lut[64:192, 1] = 255
                lut[192:, 1] = numpy.arange(252, -1, -4, dtype=numpy.uint8)
                # Blue
                lut[:64, 2] = 255
                lut[64:128, 2] = numpy.arange(254, 0, -4, dtype=numpy.uint8)
            else:
                raise RuntimeError("Built-in colormap not implemented")
            return lut

        elif description.source == "resource":
            # Load colormap LUT
            colors = numpy.load(_resource_filename("gui/colormaps/%s.npy" % name))
            # Convert to uint8 and add alpha channel
            lut = array_to_rgba8888(colors)
            return lut

        else:
            raise RuntimeError("Internal LUT source '%s' unsupported" % description.source)

    raise ValueError("Unknown colormap '%s'" % name)


def register_colormap(name, lut, cursor_color='#000000'):
    """Register a custom colormap LUT

    It can override existing LUT names.

    :param str name: Name of the LUT as defined to configure colormaps
    :param numpy.ndarray lut: The custom LUT to register.
            Nx3 or Nx4 numpy array of RGB(A) colors,
            either uint8 or float in [0, 1].
    :param str cursor_color: Color used to display overlay over images using
        colormap with this LUT.
    """
    description = _LUT_DESCRIPTION('user', cursor_color)
    colors = array_to_rgba8888(lut)
    _AVAILABLE_LUTS[name] = description

    # Register the cache as the LUT was already loaded
    _COLORMAP_CACHE[name] = colors


def get_registered_colormaps():
    """Returns currently registered colormap names"""
    return tuple(_AVAILABLE_LUTS.keys())


def get_colormap_cursor_color(name):
    """Get a color suitable for overlay over a colormap.

    :param str name: The name of the colormap.
    :return: Name of the color.
    :rtype: str
    """
    description = _AVAILABLE_LUTS.get(name, None)
    if description is not None:
        color = description.cursor_color
        if color is not None:
            return color
    return 'black'


def get_colormap_lut(name):
    """Returns the color LUT corresponding to a colormap name

    :param str name: Name of the colormap to load
    :returns: Corresponding table of colors
    :rtype: numpy.ndarray
    :raise ValueError: If no colormap corresponds to name
    """
    name = str(name)
    if name not in _COLORMAP_CACHE:
        lut = _create_colormap_lut(name)
        _COLORMAP_CACHE[name] = lut
    return _COLORMAP_CACHE[name]


# Normalizations

class _NormalizationMixIn:
    """Colormap normalization mix-in class"""

    DEFAULT_RANGE = 0, 1
    """Fallback for (vmin, vmax)"""

    def is_valid(self, value):
        """Check if a value is in the valid range for this normalization.

        Override in subclass.

        :param Union[float,numpy.ndarray] value:
        :rtype: Union[bool,numpy.ndarray]
        """
        if isinstance(value, collections.abc.Iterable):
            return numpy.ones_like(value, dtype=numpy.bool_)
        else:
            return True

    def autoscale(self, data, mode):
        """Returns range for given data and autoscale mode.

        :param Union[None,numpy.ndarray] data:
        :param str mode: Autoscale mode: 'minmax' or 'stddev3'
        :returns: Range as (min, max)
        :rtype: Tuple[float,float]
        """
        data = None if data is None else numpy.array(data, copy=False)
        if data is None or data.size == 0:
            return self.DEFAULT_RANGE

        if mode == "minmax":
            vmin, vmax = self.autoscale_minmax(data)
        elif mode == "stddev3":
            dmin, dmax = self.autoscale_minmax(data)
            stdmin, stdmax = self.autoscale_mean3std(data)
            if dmin is None:
                vmin = stdmin
            elif stdmin is None:
                vmin = dmin
            else:
                vmin = max(dmin, stdmin)

            if dmax is None:
                vmax = stdmax
            elif stdmax is None:
                vmax = dmax
            else:
                vmax = min(dmax, stdmax)

        else:
            raise ValueError('Unsupported mode: %s' % mode)

        # Check returned range and handle fallbacks
        if vmin is None or not numpy.isfinite(vmin):
            vmin = self.DEFAULT_RANGE[0]
        if vmax is None or not numpy.isfinite(vmax):
            vmax = self.DEFAULT_RANGE[1]
        if vmax < vmin:
            vmax = vmin
        return float(vmin), float(vmax)

    def autoscale_minmax(self, data):
        """Autoscale using min/max

        :param numpy.ndarray data:
        :returns: (vmin, vmax)
        :rtype: Tuple[float,float]
        """
        data = data[self.is_valid(data)]
        if data.size == 0:
            return None, None
        result = _min_max(data, min_positive=False, finite=True)
        return result.minimum, result.maximum

    def autoscale_mean3std(self, data):
        """Autoscale using mean+/-3std

        This implementation only works for normalization that do NOT
        use the data range.
        Override this method for normalization using the range.

        :param numpy.ndarray data:
        :returns: (vmin, vmax)
        :rtype: Tuple[float,float]
        """
        # Use [0, 1] as data range for normalization not using range
        normdata = self.apply(data, 0., 1.)
        if normdata.dtype.kind == 'f':  # Replaces inf by NaN
            normdata[numpy.isfinite(normdata) == False] = numpy.nan
        if normdata.size == 0:  # Fallback
            return None, None

        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=RuntimeWarning)
            # Ignore nanmean "Mean of empty slice" warning and
            # nanstd "Degrees of freedom <= 0 for slice" warning
            mean, std = numpy.nanmean(normdata), numpy.nanstd(normdata)

        return self.revert(mean - 3 * std, 0., 1.), self.revert(mean + 3 * std, 0., 1.)


class _LinearNormalizationMixIn(_NormalizationMixIn):
    """Colormap normalization mix-in class specific to autoscale taken from initial range"""

    def autoscale_mean3std(self, data):
        """Autoscale using mean+/-3std

        Do the autoscale on the data itself, not the normalized data.

        :param numpy.ndarray data:
        :returns: (vmin, vmax)
        :rtype: Tuple[float,float]
        """
        if data.dtype.kind == 'f':  # Replaces inf by NaN
            data = numpy.array(data, copy=True)  # Work on a copy
            data[numpy.isfinite(data) == False] = numpy.nan
        if data.size == 0:  # Fallback
            return None, None
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=RuntimeWarning)
            # Ignore nanmean "Mean of empty slice" warning and
            # nanstd "Degrees of freedom <= 0 for slice" warning
            mean, std = numpy.nanmean(data), numpy.nanstd(data)
        return mean - 3 * std, mean + 3 * std


class LinearNormalization(_colormap.LinearNormalization, _LinearNormalizationMixIn):
    """Linear normalization"""
    def __init__(self):
        _colormap.LinearNormalization.__init__(self)
        _LinearNormalizationMixIn.__init__(self)


class LogarithmicNormalization(_colormap.LogarithmicNormalization, _NormalizationMixIn):
    """Logarithm normalization"""

    DEFAULT_RANGE = 1, 10

    def __init__(self):
        _colormap.LogarithmicNormalization.__init__(self)
        _NormalizationMixIn.__init__(self)

    def is_valid(self, value):
        return value > 0.

    def autoscale_minmax(self, data):
        result = _min_max(data, min_positive=True, finite=True)
        return result.min_positive, result.maximum


class SqrtNormalization(_colormap.SqrtNormalization, _NormalizationMixIn):
    """Square root normalization"""

    DEFAULT_RANGE = 0, 1

    def __init__(self):
        _colormap.SqrtNormalization.__init__(self)
        _NormalizationMixIn.__init__(self)

    def is_valid(self, value):
        return value >= 0.


class GammaNormalization(_colormap.PowerNormalization, _LinearNormalizationMixIn):
    """Gamma correction normalization:

    Linear normalization to [0, 1] followed by power normalization.

    :param gamma: Gamma correction factor
    """
    def __init__(self, gamma):
        _colormap.PowerNormalization.__init__(self, gamma)
        _LinearNormalizationMixIn.__init__(self)


# Backward compatibility
PowerNormalization = GammaNormalization


class ArcsinhNormalization(_colormap.ArcsinhNormalization, _NormalizationMixIn):
    """Inverse hyperbolic sine normalization"""

    def __init__(self):
        _colormap.ArcsinhNormalization.__init__(self)
        _NormalizationMixIn.__init__(self)


# Colormap function

_BASIC_NORMALIZATIONS = {
    "linear": LinearNormalization(),
    "log": LogarithmicNormalization(),
    "sqrt": SqrtNormalization(),
    "arcsinh": ArcsinhNormalization(),
}

_DEFAULT_NAN_COLOR = 255, 255, 255, 0

def apply_colormap(data,
                   colormap: str,
                   norm: str="linear",
                   autoscale: str="minmax",
                   vmin=None,
                   vmax=None,
                   gamma=1.0):
    """Apply colormap to data with given normalization and autoscale.

    :param numpy.ndarray data: Data on which to apply the colormap
    :param str colormap: Name of the colormap to use
    :param str norm: Normalization to use
    :param str autoscale: Autoscale mode: "minmax" (default) or "stddev3"
    :param vmin: Lower bound, None (default) to autoscale
    :param vmax: Upper bound, None (default) to autoscale
    :param float gamma:
        Gamma correction parameter (used only for "gamma" normalization)
    :returns: Array of colors
    """
    colors = get_colormap_lut(colormap)

    if norm == "gamma":
        normalizer = GammaNormalization(gamma)
    else:
        normalizer = _BASIC_NORMALIZATIONS[norm]

    if vmin is None or vmax is None:
        auto_vmin, auto_vmax = normalizer.autoscale(data, autoscale)
        if vmin is None:  # Set vmin respecting provided vmax
            vmin = auto_vmin if vmax is None else min(auto_vmin, vmax)
        if vmax is None:
            vmax = max(auto_vmax, vmin)  # Handle max_ <= 0 for log scale

    return _colormap.cmap(
        data,
        colors,
        vmin,
        vmax,
        normalizer,
        _DEFAULT_NAN_COLOR,
    )



# /*##########################################################################
#
# Copyright (c) 2015-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""QWidget displaying an overview of a 2D plot.

This shows the available range of the data, and the current location of the
plot view.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "22/02/2021"

import logging
import weakref
from ... import qt
from ...utils import LockReentrant

_logger = logging.getLogger(__name__)


class _DraggableRectItem(qt.QGraphicsRectItem):
    """RectItem which signals its change through visibleRectDragged."""
    def __init__(self, *args, **kwargs):
        super(_DraggableRectItem, self).__init__(
            *args, **kwargs)

        self._previousCursor = None
        self.setFlag(qt.QGraphicsItem.ItemIsMovable)
        self.setFlag(qt.QGraphicsItem.ItemSendsGeometryChanges)
        self.setAcceptHoverEvents(True)
        self._ignoreChange = False
        self._constraint = 0, 0, 0, 0

    def setConstraintRect(self, left, top, width, height):
        """Set the constraint rectangle for dragging.

        The coordinates are in the _DraggableRectItem coordinate system.

        This constraint only applies to modification through interaction
        (i.e., this constraint is not applied to change through API).

        If the _DraggableRectItem is smaller than the constraint rectangle,
        the _DraggableRectItem remains within the constraint rectangle.
        If the _DraggableRectItem is wider than the constraint rectangle,
        the constraint rectangle remains within the _DraggableRectItem.
        """
        self._constraint = left, left + width, top, top + height

    def setPos(self, *args, **kwargs):
        """Overridden to ignore changes from API in itemChange."""
        self._ignoreChange = True
        super(_DraggableRectItem, self).setPos(*args, **kwargs)
        self._ignoreChange = False

    def moveBy(self, *args, **kwargs):
        """Overridden to ignore changes from API in itemChange."""
        self._ignoreChange = True
        super(_DraggableRectItem, self).moveBy(*args, **kwargs)
        self._ignoreChange = False

    def itemChange(self, change, value):
        """Callback called before applying changes to the item."""
        if (change == qt.QGraphicsItem.ItemPositionChange and
                not self._ignoreChange):
            # Makes sure that the visible area is in the data
            # or that data is in the visible area if area is too wide
            x, y = value.x(), value.y()
            xMin, xMax, yMin, yMax = self._constraint

            if self.rect().width() <= (xMax - xMin):
                if x < xMin:
                    value.setX(xMin)
                elif x > xMax - self.rect().width():
                    value.setX(xMax - self.rect().width())
            else:
                if x > xMin:
                    value.setX(xMin)
                elif x < xMax - self.rect().width():
                    value.setX(xMax - self.rect().width())

            if self.rect().height() <= (yMax - yMin):
                if y < yMin:
                    value.setY(yMin)
                elif y > yMax - self.rect().height():
                    value.setY(yMax - self.rect().height())
            else:
                if y > yMin:
                    value.setY(yMin)
                elif y < yMax - self.rect().height():
                    value.setY(yMax - self.rect().height())

            if self.pos() != value:
                # Notify change through signal
                views = self.scene().views()
                assert len(views) == 1
                views[0].visibleRectDragged.emit(
                    value.x() + self.rect().left(),
                    value.y() + self.rect().top(),
                    self.rect().width(),
                    self.rect().height())

            return value

        return super(_DraggableRectItem, self).itemChange(
            change, value)

    def hoverEnterEvent(self, event):
        """Called when the mouse enters the rectangle area"""
        self._previousCursor = self.cursor()
        self.setCursor(qt.Qt.OpenHandCursor)

    def hoverLeaveEvent(self, event):
        """Called when the mouse leaves the rectangle area"""
        if self._previousCursor is not None:
            self.setCursor(self._previousCursor)
            self._previousCursor = None


class RadarView(qt.QGraphicsView):
    """Widget presenting a synthetic view of a 2D area and
    the current visible area.

    Coordinates are as in QGraphicsView:
    x goes from left to right and y goes from top to bottom.
    This widget preserves the aspect ratio of the areas.

    The 2D area and the visible area can be set with :meth:`setDataRect`
    and :meth:`setVisibleRect`.
    When the visible area has been dragged by the user, its new position
    is signaled by the *visibleRectDragged* signal.

    It is possible to invert the direction of the axes by using the
    :meth:`scale` method of QGraphicsView.
    """

    visibleRectDragged = qt.Signal(float, float, float, float)
    """Signals that the visible rectangle has been dragged.

    It provides: left, top, width, height in data coordinates.
    """

    _DATA_PEN = qt.QPen(qt.QColor('white'))
    _DATA_BRUSH = qt.QBrush(qt.QColor('light gray'))
    _ACTIVEDATA_PEN = qt.QPen(qt.QColor('black'))
    _ACTIVEDATA_BRUSH = qt.QBrush(qt.QColor('transparent'))
    _ACTIVEDATA_PEN.setWidth(2)
    _ACTIVEDATA_PEN.setCosmetic(True)
    _VISIBLE_PEN = qt.QPen(qt.QColor('blue'))
    _VISIBLE_PEN.setWidth(2)
    _VISIBLE_PEN.setCosmetic(True)
    _VISIBLE_BRUSH = qt.QBrush(qt.QColor(0, 0, 0, 0))
    _TOOLTIP = 'Radar View:\nRed contour: Visible area\nGray area: The image'

    _PIXMAP_SIZE = 256

    def __init__(self, parent=None):
        self.__plotRef = None
        self._scene = qt.QGraphicsScene()
        self._dataRect = self._scene.addRect(0, 0, 1, 1,
                                             self._DATA_PEN,
                                             self._DATA_BRUSH)
        self._imageRect = self._scene.addRect(0, 0, 1, 1,
                                              self._ACTIVEDATA_PEN,
                                              self._ACTIVEDATA_BRUSH)
        self._imageRect.setVisible(False)
        self._scatterRect = self._scene.addRect(0, 0, 1, 1,
                                                self._ACTIVEDATA_PEN,
                                                self._ACTIVEDATA_BRUSH)
        self._scatterRect.setVisible(False)
        self._curveRect = self._scene.addRect(0, 0, 1, 1,
                                              self._ACTIVEDATA_PEN,
                                              self._ACTIVEDATA_BRUSH)
        self._curveRect.setVisible(False)

        self._visibleRect = _DraggableRectItem(0, 0, 1, 1)
        self._visibleRect.setPen(self._VISIBLE_PEN)
        self._visibleRect.setBrush(self._VISIBLE_BRUSH)
        self._scene.addItem(self._visibleRect)

        super(RadarView, self).__init__(self._scene, parent)
        self.setHorizontalScrollBarPolicy(qt.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(qt.Qt.ScrollBarAlwaysOff)
        self.setFocusPolicy(qt.Qt.NoFocus)
        self.setStyleSheet('border: 0px')
        self.setToolTip(self._TOOLTIP)

        self.__reentrant = LockReentrant()
        self.visibleRectDragged.connect(self._viewRectDragged)

        self.__timer = qt.QTimer(self)
        self.__timer.timeout.connect(self._updateDataContent)

    def sizeHint(self):
        # """Overridden to avoid sizeHint to depend on content size."""
        return self.minimumSizeHint()

    def wheelEvent(self, event):
        # """Overridden to disable vertical scrolling with wheel."""
        event.ignore()

    def resizeEvent(self, event):
        # """Overridden to fit current content to new size."""
        self.fitInView(self._scene.itemsBoundingRect(), qt.Qt.KeepAspectRatio)
        super(RadarView, self).resizeEvent(event)

    def setDataRect(self, left, top, width, height):
        """Set the bounds of the data rectangular area.

        This sets the coordinate system.
        """
        self._dataRect.setRect(left, top, width, height)
        self._visibleRect.setConstraintRect(left, top, width, height)
        self.fitInView(self._scene.itemsBoundingRect(), qt.Qt.KeepAspectRatio)

    def setVisibleRect(self, left, top, width, height):
        """Set the visible rectangular area.

        The coordinates are relative to the data rect.
        """
        self.__visibleRect = left, top, width, height
        self._visibleRect.setRect(0, 0, width, height)
        self._visibleRect.setPos(left, top)
        self.fitInView(self._scene.itemsBoundingRect(), qt.Qt.KeepAspectRatio)

    def __setVisibleRectFromPlot(self, plot):
        """Update radar view visible area.

        Takes care of y coordinate conversion.
        """
        xMin, xMax = plot.getXAxis().getLimits()
        yMin, yMax = plot.getYAxis().getLimits()
        self.setVisibleRect(xMin, yMin, xMax - xMin, yMax - yMin)

    def getPlotWidget(self):
        """Returns the connected plot

        :rtype: Union[None,PlotWidget]
        """
        if self.__plotRef is None:
            return None
        plot = self.__plotRef()
        if plot is None:
            self.__plotRef = None
        return plot

    def setPlotWidget(self, plot):
        """Set the PlotWidget this radar view connects to.

        As result `setDataRect` and `setVisibleRect` will be called
        automatically.

        :param Union[None,PlotWidget] plot:
        """
        previousPlot = self.getPlotWidget()
        if previousPlot is not None:  # Disconnect previous plot
            plot.getXAxis().sigLimitsChanged.disconnect(self._xLimitChanged)
            plot.getYAxis().sigLimitsChanged.disconnect(self._yLimitChanged)
            plot.getYAxis().sigInvertedChanged.disconnect(self._updateYAxisInverted)

        # Reset plot and timer
        # FIXME: It would be good to clean up the display here
        self.__plotRef = None
        self.__timer.stop()

        if plot is not None:  # Connect new plot
            self.__plotRef = weakref.ref(plot)
            plot.getXAxis().sigLimitsChanged.connect(self._xLimitChanged)
            plot.getYAxis().sigLimitsChanged.connect(self._yLimitChanged)
            plot.getYAxis().sigInvertedChanged.connect(self._updateYAxisInverted)
            self.__setVisibleRectFromPlot(plot)
            self._updateYAxisInverted()
            self.__timer.start(500)

    def _xLimitChanged(self, vmin, vmax):
        plot = self.getPlotWidget()
        self.__setVisibleRectFromPlot(plot)

    def _yLimitChanged(self, vmin, vmax):
        plot = self.getPlotWidget()
        self.__setVisibleRectFromPlot(plot)

    def _updateYAxisInverted(self, inverted=None):
        """Sync radar view axis orientation."""
        plot = self.getPlotWidget()
        if inverted is None:
            # Do not perform this when called from plot signal
            inverted = plot.getYAxis().isInverted()
        # Use scale to invert radarView
        # RadarView default Y direction is from top to bottom
        # As opposed to Plot. So invert RadarView when Plot is NOT inverted.
        self.resetTransform()
        if not inverted:
            self.scale(1., -1.)
        self.update()

    def _viewRectDragged(self, left, top, width, height):
        """Slot for radar view visible rectangle changes."""
        plot = self.getPlotWidget()
        if plot is None:
            return

        if self.__reentrant.locked():
            return

        with self.__reentrant:
            plot.setLimits(left, left + width, top, top + height)

    def _updateDataContent(self):
        """Update the content to the current data content"""
        plot = self.getPlotWidget()
        if plot is None:
            return
        ranges = plot.getDataRange()
        xmin, xmax = ranges.x if ranges.x is not None else (0, 0)
        ymin, ymax = ranges.y if ranges.y is not None else (0, 0)
        self.setDataRect(xmin, ymin, xmax - xmin, ymax - ymin)

        self.__updateItem(self._imageRect, plot.getActiveImage())
        self.__updateItem(self._scatterRect, plot.getActiveScatter())
        self.__updateItem(self._curveRect, plot.getActiveCurve())

    def __updateItem(self, rect, item):
        """Sync rect with item bounds

        :param QGraphicsRectItem rect:
        :param Item item:
        """
        if item is None:
            rect.setVisible(False)
            return
        ranges = item._getBounds()
        if ranges is None:
            rect.setVisible(False)
            return
        xmin, xmax, ymin, ymax = ranges
        width = xmax - xmin
        height = ymax - ymin
        rect.setRect(xmin, ymin, width, height)
        rect.setVisible(True)



# /*##########################################################################
#
# Copyright (c) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Utilitary functions dealing with numbers.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "05/06/2018"

import numpy
import re
import logging


_logger = logging.getLogger(__name__)


_biggest_float = None

if hasattr(numpy, "longdouble"):
    finfo = numpy.finfo(numpy.longdouble)
    # The bit for sign is missing here
    bits = finfo.nexp + finfo.nmant
    if bits > 64:
        _biggest_float = numpy.longdouble
        # From bigger to smaller
        _float_types = (numpy.longdouble, numpy.float64, numpy.float32, numpy.float16)
if _biggest_float is None:
    _biggest_float = numpy.float64
    # From bigger to smaller
    _float_types = (numpy.float64, numpy.float32, numpy.float16)


_parse_numeric_value = re.compile(r"^\s*[-+]?0*(\d+?)?(?:\.(\d+))?(?:[eE]([-+]?\d+))?\s*$")


def is_longdouble_64bits():
    """Returns true if the system uses floating-point 64bits for it's
    long double type.

    .. note:: Comparing `numpy.longdouble` and `numpy.float64` on Windows is not
        possible (or at least not will all the numpy version)
    """
    return _biggest_float == numpy.float64


def min_numerical_convertible_type(string, check_accuracy=True):
    """
    Parse the string and try to return the smallest numerical type to use for
    a safe conversion. It has some known issues: precission loss.

    :param str string: Representation of a float/integer with text
    :param bool check_accuracy: If true, a warning is pushed on the logger
        in case there is a loss of accuracy.
    :raise ValueError: When the string is not a numerical value
    :retrun: A numpy numerical type
    """
    if string == "":
        raise ValueError("Not a numerical value")
    match = _parse_numeric_value.match(string)
    if match is None:
        raise ValueError("Not a numerical value")
    number, decimal, exponent = match.groups()

    if decimal is None and exponent is None:
        # It's an integer
        # TODO: We could find the int type without converting the number
        value = int(string)
        return numpy.min_scalar_type(value).type

    # Try floating-point
    try:
        value = _biggest_float(string)
    except ValueError:
        raise ValueError("Not a numerical value")

    if number is None:
        number = ""
    if decimal is None:
        decimal = ""
    if exponent is None:
        exponent = "0"

    nb_precision_digits = int(exponent) - len(decimal) - 1
    precision = _biggest_float(10) ** nb_precision_digits * 1.2
    previous_type = _biggest_float
    for numpy_type in _float_types:
        if numpy_type == _biggest_float:
            # value was already casted using the bigger type
            continue
        reduced_value = numpy_type(value)
        if not numpy.isfinite(reduced_value):
            break
        # numpy isclose(atol=is not accurate enough)
        diff = value - reduced_value
        # numpy 1.8.2 looks to do the substraction using float64...
        # we lose precision here
        diff = numpy.abs(diff)
        if diff > precision:
            break
        previous_type = numpy_type

    # It's the smaller float type which fit with enougth precision
    numpy_type = previous_type

    if check_accuracy and numpy_type == _biggest_float:
        # Check the precision using the original string
        expected = number + decimal
        # This format the number without python convertion
        try:
            result = numpy.array2string(value, precision=len(number) + len(decimal), floatmode="fixed")
        except TypeError:
            # numpy 1.8.2 do not have floatmode argument
            _logger.warning("Not able to check accuracy of the conversion of '%s' using %s", string, _biggest_float)
            return numpy_type

        result = result.replace(".", "").replace("-", "")
        if not result.startswith(expected):
            _logger.warning("Not able to convert '%s' using %s without losing precision", string, _biggest_float)

    return numpy_type



# /*##########################################################################
#
# Copyright (c) 2014-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
This module provides convenient classes and functions for OpenGL rendering.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "03/04/2017"


import numpy

from ...._glutils import gl


def buildFillMaskIndices(nIndices, dtype=None):
    """Returns triangle strip indices for rendering a filled polygon mask

    :param int nIndices: Number of points
    :param Union[numpy.dtype,None] dtype:
       If specified the dtype of the returned indices array
    :return: 1D array of indices constructing a triangle strip
    :rtype: numpy.ndarray
    """
    if dtype is None:
        if nIndices <= numpy.iinfo(numpy.uint16).max + 1:
            dtype = numpy.uint16
        else:
            dtype = numpy.uint32

    lastIndex = nIndices - 1
    splitIndex = lastIndex // 2 + 1
    indices = numpy.empty(nIndices, dtype=dtype)
    indices[::2] = numpy.arange(0, splitIndex, step=1, dtype=dtype)
    indices[1::2] = numpy.arange(lastIndex, splitIndex - 1, step=-1,
                                 dtype=dtype)
    return indices


class FilledShape2D(object):
    _NO_HATCH = 0
    _HATCH_STEP = 20

    def __init__(self, points, style='solid', color=(0., 0., 0., 1.)):
        self.vertices = numpy.array(points, dtype=numpy.float32, copy=False)
        self._indices = buildFillMaskIndices(len(self.vertices))

        tVertex = numpy.transpose(self.vertices)
        xMin, xMax = min(tVertex[0]), max(tVertex[0])
        yMin, yMax = min(tVertex[1]), max(tVertex[1])
        self.bboxVertices = numpy.array(((xMin, yMin), (xMin, yMax),
                                         (xMax, yMin), (xMax, yMax)),
                                        dtype=numpy.float32)
        self._xMin, self._xMax = xMin, xMax
        self._yMin, self._yMax = yMin, yMax

        self.style = style
        self.color = color

    def render(self, posAttrib, colorUnif, hatchStepUnif):
        assert self.style in ('hatch', 'solid')
        gl.glUniform4f(colorUnif, *self.color)
        step = self._HATCH_STEP if self.style == 'hatch' else self._NO_HATCH
        gl.glUniform1i(hatchStepUnif, step)

        # Prepare fill mask
        gl.glEnableVertexAttribArray(posAttrib)
        gl.glVertexAttribPointer(posAttrib,
                                 2,
                                 gl.GL_FLOAT,
                                 gl.GL_FALSE,
                                 0, self.vertices)

        gl.glEnable(gl.GL_STENCIL_TEST)
        gl.glStencilMask(1)
        gl.glStencilFunc(gl.GL_ALWAYS, 1, 1)
        gl.glStencilOp(gl.GL_INVERT, gl.GL_INVERT, gl.GL_INVERT)
        gl.glColorMask(gl.GL_FALSE, gl.GL_FALSE, gl.GL_FALSE, gl.GL_FALSE)
        gl.glDepthMask(gl.GL_FALSE)

        gl.glDrawElements(gl.GL_TRIANGLE_STRIP, len(self._indices),
                          gl.GL_UNSIGNED_SHORT, self._indices)

        gl.glStencilFunc(gl.GL_EQUAL, 1, 1)
        # Reset stencil while drawing
        gl.glStencilOp(gl.GL_ZERO, gl.GL_ZERO, gl.GL_ZERO)
        gl.glColorMask(gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE)
        gl.glDepthMask(gl.GL_TRUE)

        gl.glVertexAttribPointer(posAttrib,
                                 2,
                                 gl.GL_FLOAT,
                                 gl.GL_FALSE,
                                 0, self.bboxVertices)
        gl.glDrawArrays(gl.GL_TRIANGLE_STRIP, 0, len(self.bboxVertices))

        gl.glDisable(gl.GL_STENCIL_TEST)


# matrix ######################################################################

def mat4Ortho(left, right, bottom, top, near, far):
    """Orthographic projection matrix (row-major)"""
    return numpy.array((
        (2./(right - left), 0., 0., -(right+left)/float(right-left)),
        (0., 2./(top - bottom), 0., -(top+bottom)/float(top-bottom)),
        (0., 0., -2./(far-near),    -(far+near)/float(far-near)),
        (0., 0., 0., 1.)), dtype=numpy.float64)


def mat4Translate(x=0., y=0., z=0.):
    """Translation matrix (row-major)"""
    return numpy.array((
        (1., 0., 0., x),
        (0., 1., 0., y),
        (0., 0., 1., z),
        (0., 0., 0., 1.)), dtype=numpy.float64)


def mat4Scale(sx=1., sy=1., sz=1.):
    """Scale matrix (row-major)"""
    return numpy.array((
        (sx, 0., 0., 0.),
        (0., sy, 0., 0.),
        (0., 0., sz, 0.),
        (0., 0., 0., 1.)), dtype=numpy.float64)


def mat4Identity():
    """Identity matrix"""
    return numpy.array((
        (1., 0., 0., 0.),
        (0., 1., 0., 0.),
        (0., 0., 1., 0.),
        (0., 0., 0., 1.)), dtype=numpy.float64)



# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Utilities for writing tests.

- :class:`ParametricTestCase` provides a :meth:`TestCase.subTest` replacement
  for Python < 3.4
- :class:`LoggingValidator` with context or the :func:`validate_logging`
  decorator enables testing the number of logging messages of different levels.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "26/01/2018"


import contextlib
import functools
import logging
import sys
import unittest
from . import deprecation


_logger = logging.getLogger(__name__)


if sys.hexversion >= 0x030400F0:  # Python >= 3.4
    class ParametricTestCase(unittest.TestCase):
        pass
else:
    class ParametricTestCase(unittest.TestCase):
        """TestCase with subTest support for Python < 3.4.

        Add subTest method to support parametric tests.
        API is the same, but behavior differs:
        If a subTest fails, the following ones are not run.
        """

        _subtest_msg = None  # Class attribute to provide a default value

        @contextlib.contextmanager
        def subTest(self, msg=None, **params):
            """Use as unittest.TestCase.subTest method in Python >= 3.4."""
            # Format arguments as: '[msg] (key=value, ...)'
            param_str = ', '.join(['%s=%s' % (k, v) for k, v in params.items()])
            self._subtest_msg = '[%s] (%s)' % (msg or '', param_str)
            yield
            self._subtest_msg = None

        def shortDescription(self):
            short_desc = super(ParametricTestCase, self).shortDescription()
            if self._subtest_msg is not None:
                # Append subTest message to shortDescription
                short_desc = ' '.join(
                    [msg for msg in (short_desc, self._subtest_msg) if msg])

            return short_desc if short_desc else None


def parameterize(test_case_class, *args, **kwargs):
    """Create a suite containing all tests taken from the given
    subclass, passing them the parameters.

    .. code-block:: python

        class TestParameterizedCase(unittest.TestCase):
            def __init__(self, methodName='runTest', foo=None):
                unittest.TestCase.__init__(self, methodName)
                self.foo = foo

        def suite():
            testSuite = unittest.TestSuite()
            testSuite.addTest(parameterize(TestParameterizedCase, foo=10))
            testSuite.addTest(parameterize(TestParameterizedCase, foo=50))
            return testSuite
    """
    test_loader = unittest.TestLoader()
    test_names = test_loader.getTestCaseNames(test_case_class)
    suite = unittest.TestSuite()
    for name in test_names:
        suite.addTest(test_case_class(name, *args, **kwargs))
    return suite


class LoggingRuntimeError(RuntimeError):
    """Raised when the `LoggingValidator` fails"""

    def __init__(self, msg, records):
        super(LoggingRuntimeError, self).__init__(msg)
        self.records = records

    def __str__(self):
        return super(LoggingRuntimeError, self).__str__() + " -> " + str(self.records)


class LoggingValidator(logging.Handler):
    """Context checking the number of logging messages from a specified Logger.

    It disables propagation of logging message while running.

    This is meant to be used as a with statement, for example:

    >>> with LoggingValidator(logger, error=2, warning=0):
    >>>     pass  # Run tests here expecting 2 ERROR and no WARNING from logger
    ...

    :param logger: Name or instance of the logger to test.
                   (Default: root logger)
    :type logger: str or :class:`logging.Logger`
    :param int critical: Expected number of CRITICAL messages.
                         Default: Do not check.
    :param int error: Expected number of ERROR messages.
                      Default: Do not check.
    :param int warning: Expected number of WARNING messages.
                        Default: Do not check.
    :param int info: Expected number of INFO messages.
                     Default: Do not check.
    :param int debug: Expected number of DEBUG messages.
                      Default: Do not check.
    :param int notset: Expected number of NOTSET messages.
                       Default: Do not check.
    :raises RuntimeError: If the message counts are the expected ones.
    """

    def __init__(self, logger=None, critical=None, error=None,
                 warning=None, info=None, debug=None, notset=None):
        if logger is None:
            logger = logging.getLogger()
        elif not isinstance(logger, logging.Logger):
            logger = logging.getLogger(logger)
        self.logger = logger

        self.records = []

        self.expected_count_by_level = {
            logging.CRITICAL: critical,
            logging.ERROR: error,
            logging.WARNING: warning,
            logging.INFO: info,
            logging.DEBUG: debug,
            logging.NOTSET: notset
        }

        self._expected_count = sum([v for k, v in self.expected_count_by_level.items() if v is not None])
        """Amount of any logging expected"""

        super(LoggingValidator, self).__init__()

    def __enter__(self):
        """Context (i.e., with) support"""
        self.records = []  # Reset recorded LogRecords
        self.logger.addHandler(self)
        self.logger.propagate = False
        # ensure no log message is ignored
        self.entry_level = self.logger.level * 1
        self.logger.setLevel(logging.DEBUG)
        self.entry_disabled = self.logger.disabled
        self.logger.disabled = False
        return self

    def can_be_checked(self):
        """Returns True if this listener have received enough messages to
        be valid, and then checked.

        This can be useful for asynchronous wait of messages. It allows process
        an early break, instead of waiting much time in an active loop.
        """
        return len(self.records) >= self._expected_count

    def get_count_by_level(self):
        """Returns the current message count by level.
        """
        count = {
            logging.CRITICAL: 0,
            logging.ERROR: 0,
            logging.WARNING: 0,
            logging.INFO: 0,
            logging.DEBUG: 0,
            logging.NOTSET: 0
        }
        for record in self.records:
            level = record.levelno
            if level in count:
                count[level] = count[level] + 1
        return count

    def __exit__(self, exc_type, exc_value, traceback):
        """Context (i.e., with) support"""
        self.logger.removeHandler(self)
        self.logger.propagate = True
        self.logger.setLevel(self.entry_level)
        self.logger.disabled = self.entry_disabled

        count_by_level = self.get_count_by_level()

        # Remove keys which does not matter
        ignored = [r for r, v in self.expected_count_by_level.items() if v is None]
        expected_count_by_level = dict(self.expected_count_by_level)
        for i in ignored:
            del count_by_level[i]
            del expected_count_by_level[i]

        if count_by_level != expected_count_by_level:
            # Re-send record logs through logger as they where masked
            # to help debug
            message = ""
            for level in count_by_level.keys():
                if message != "":
                    message += ", "
                count = count_by_level[level]
                expected_count = expected_count_by_level[level]
                message += "%d %s (got %d)" % (expected_count, logging.getLevelName(level), count)

            raise LoggingRuntimeError(
                'Expected %s' % message, records=list(self.records))

    def emit(self, record):
        """Override :meth:`logging.Handler.emit`"""
        self.records.append(record)


def validate_logging(logger=None, critical=None, error=None,
                     warning=None, info=None, debug=None, notset=None):
    """Decorator checking number of logging messages.

    Propagation of logging messages is disabled by this decorator.

    In case the expected number of logging messages is not found, it raises
    a RuntimeError.

    >>> class Test(unittest.TestCase):
    ...     @validate_logging('module_logger_name', error=2, warning=0)
    ...     def test(self):
    ...         pass  # Test expecting 2 ERROR and 0 WARNING messages

    :param logger: Name or instance of the logger to test.
                   (Default: root logger)
    :type logger: str or :class:`logging.Logger`
    :param int critical: Expected number of CRITICAL messages.
                         Default: Do not check.
    :param int error: Expected number of ERROR messages.
                      Default: Do not check.
    :param int warning: Expected number of WARNING messages.
                        Default: Do not check.
    :param int info: Expected number of INFO messages.
                     Default: Do not check.
    :param int debug: Expected number of DEBUG messages.
                      Default: Do not check.
    :param int notset: Expected number of NOTSET messages.
                       Default: Do not check.
    """
    def decorator(func):
        test_context = LoggingValidator(
            logger, critical, error, warning, info, debug, notset)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with test_context:
                result = func(*args, **kwargs)
            return result
        return wrapper
    return decorator


# Backward compatibility
class TestLogging(LoggingValidator):
    def __init__(self, *args, **kwargs):
        deprecation.deprecated_warning(
            "Class",
            "TestLogging",
            since_version="1.0.0",
            replacement="LoggingValidator")
        super().__init__(*args, **kwargs)


@deprecation.deprecated(since_version="1.0.0", replacement="validate_logging")
def test_logging(*args, **kwargs):
    return validate_logging(*args, **kwargs)


# Simulate missing library context
class EnsureImportError(object):
    """This context manager allows to simulate the unavailability
    of a library, even if it is actually available. It ensures that
    an ImportError is raised if the code inside the context tries to
    import the module.

    It can be used to test that a correct fallback library is used,
    or that the expected error code is returned.

    Trivial example::

        from silx.utils.testutils import EnsureImportError

        with EnsureImportError("h5py"):
            try:
                import h5py
            except ImportError:
                print("Good")

    .. note::

        This context manager does not remove the library from the namespace,
        if it is already imported. It only ensures that any attempt to import
        it again will cause an ImportError to be raised.
    """
    def __init__(self, name):
        """

        :param str name: Name of module to be hidden (e.g. "h5py")
        """
        self.module_name = name

    def __enter__(self):
        """Simulate failed import by setting sys.modules[name]=None"""
        if self.module_name not in sys.modules:
            self._delete_on_exit = True
            self._backup = None
        else:
            self._delete_on_exit = False
            self._backup = sys.modules[self.module_name]
        sys.modules[self.module_name] = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Restore previous state"""
        if self._delete_on_exit:
            del sys.modules[self.module_name]
        else:
            sys.modules[self.module_name] = self._backup



# /*##########################################################################
# Copyright (C) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Tests for specfile wrapper"""

__authors__ = ["P. Knobel", "V.A. Sole"]
__license__ = "MIT"
__date__ = "17/01/2018"


import locale
import logging
import numpy
import os
import sys
import tempfile
import unittest

from silx.utils import testutils

from ..specfile import SpecFile, Scan
from .. import specfile


logger1 = logging.getLogger(__name__)

sftext = """#F /tmp/sf.dat
#E 1455180875
#D Thu Feb 11 09:54:35 2016
#C imaging  User = opid17
#U00 user comment first line
#U01 This is a dummy file to test SpecFile parsing
#U02
#U03 last line

#O0 Pslit HGap  MRTSlit UP  MRTSlit DOWN
#O1 Sslit1 VOff  Sslit1 HOff  Sslit1 VGap
#o0 pshg mrtu mrtd
#o2 ss1vo ss1ho ss1vg

#J0 Seconds  IA  ion.mono  Current
#J1 xbpmc2  idgap1  Inorm

#S 1  ascan  ss1vo -4.55687 -0.556875  40 0.2
#D Thu Feb 11 09:55:20 2016
#T 0.2  (Seconds)
#G0 0
#G1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
#G3 0 0 0 0 0 0 0 0 0
#G4 0
#Q
#P0 180.005 -0.66875 0.87125
#P1 14.74255 16.197579 12.238283
#UMI0     Current AutoM      Shutter
#UMI1      192.51   OFF     FE open
#UMI2 Refill in 39883 sec, Fill Mode: uniform multibunch / Message: Feb 11 08:00 Delivery:Next Refill at 21:00;
#N 4
#L first column  second column  3rd_col
-1.23 5.89  8
8.478100E+01  5 1.56
3.14 2.73 -3.14
1.2 2.3 3.4

#S 25  ascan  c3th 1.33245 1.52245  40 0.15
#D Thu Feb 11 10:00:31 2016
#P0 80.005 -1.66875 1.87125
#P1 4.74255 6.197579 2.238283
#N 5
#L column0  column1  col2  col3
0.0 0.1 0.2 0.3
1.0 1.1 1.2 1.3
2.0 2.1 2.2 2.3
3.0 3.1 3.2 3.3

#S 26  yyyyyy
#D Thu Feb 11 09:55:20 2016
#P0 80.005 -1.66875 1.87125
#P1 4.74255 6.197579 2.238283
#N 4
#L first column  second column  3rd_col
#C Sat Oct 31 15:51:47 1998.  Scan aborted after 0 points.

#F /tmp/sf.dat
#E 1455180876
#D Thu Feb 11 09:54:36 2016

#S 1 aaaaaa
#U first duplicate line
#U second duplicate line
#@MCADEV 1
#@MCA %16C
#@CHANN 3 0 2 1
#@CALIB 1 2 3
#N 3
#L uno  duo
1 2
@A 0 1 2
3 4
@A 3.1 4 5
5 6
@A 6 7.7 8
"""


loc = locale.getlocale(locale.LC_NUMERIC)
try:
    locale.setlocale(locale.LC_NUMERIC, 'de_DE.utf8')
except locale.Error:
    try_DE = False
else:
    try_DE = True
    locale.setlocale(locale.LC_NUMERIC, loc)


class TestSpecFile(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        fd, cls.fname1 = tempfile.mkstemp(text=False)
        if sys.version_info < (3, ):
            os.write(fd, sftext)
        else:
            os.write(fd, bytes(sftext, 'ascii'))
        os.close(fd)

        fd2, cls.fname2 = tempfile.mkstemp(text=False)
        if sys.version_info < (3, ):
            os.write(fd2, sftext[370:923])
        else:
            os.write(fd2, bytes(sftext[370:923], 'ascii'))
        os.close(fd2)

        fd3, cls.fname3 = tempfile.mkstemp(text=False)
        txt = sftext[371:923]
        if sys.version_info < (3, ):
            os.write(fd3, txt)
        else:
            os.write(fd3, bytes(txt, 'ascii'))
        os.close(fd3)

    @classmethod
    def tearDownClass(cls):
        os.unlink(cls.fname1)
        os.unlink(cls.fname2)
        os.unlink(cls.fname3)

    def setUp(self):
        self.sf = SpecFile(self.fname1)
        self.scan1 = self.sf[0]
        self.scan1_2 = self.sf["1.2"]
        self.scan25 = self.sf["25.1"]
        self.empty_scan = self.sf["26.1"]

        self.sf_no_fhdr = SpecFile(self.fname2)
        self.scan1_no_fhdr = self.sf_no_fhdr[0]

        self.sf_no_fhdr_crash = SpecFile(self.fname3)
        self.scan1_no_fhdr_crash = self.sf_no_fhdr_crash[0]

    def tearDown(self):
        self.sf.close()
        self.sf_no_fhdr.close()
        self.sf_no_fhdr_crash.close()

    def test_open(self):
        self.assertIsInstance(self.sf, SpecFile)
        with self.assertRaises(specfile.SfErrFileOpen):
            SpecFile("doesnt_exist.dat")

        # test filename types unicode and bytes
        if sys.version_info[0] < 3:
            try:
                SpecFile(self.fname1)
            except TypeError:
                self.fail("failed to handle filename as python2 str")
            try:
                SpecFile(unicode(self.fname1))
            except TypeError:
                self.fail("failed to handle filename as python2 unicode")
        else:
            try:
                SpecFile(self.fname1)
            except TypeError:
                self.fail("failed to handle filename as python3 str")
            try:
                SpecFile(bytes(self.fname1, 'utf-8'))
            except TypeError:
                self.fail("failed to handle filename as python3 bytes")

    def test_number_of_scans(self):
        self.assertEqual(4, len(self.sf))

    def test_list_of_scan_indices(self):
        self.assertEqual(self.sf.list(),
                         [1, 25, 26, 1])
        self.assertEqual(self.sf.keys(),
                         ["1.1", "25.1", "26.1", "1.2"])

    def test_index_number_order(self):
        self.assertEqual(self.sf.index(1, 2), 3)  # sf["1.2"]==sf[3]
        self.assertEqual(self.sf.number(1), 25)   # sf[1]==sf["25"]
        self.assertEqual(self.sf.order(3), 2)     # sf[3]==sf["1.2"]
        with self.assertRaises(specfile.SfErrScanNotFound):
            self.sf.index(3, 2)
        with self.assertRaises(specfile.SfErrScanNotFound):
            self.sf.index(99)

    def assertRaisesRegex(self, *args, **kwargs):
        # Python 2 compatibility
        if sys.version_info.major >= 3:
            return super(TestSpecFile, self).assertRaisesRegex(*args, **kwargs)
        else:
            return self.assertRaisesRegexp(*args, **kwargs)

    def test_getitem(self):
        self.assertIsInstance(self.sf[2], Scan)
        self.assertIsInstance(self.sf["1.2"], Scan)
        # int out of range
        with self.assertRaisesRegex(IndexError, 'Scan index must be in ran'):
            self.sf[107]
        # float indexing not allowed
        with self.assertRaisesRegex(TypeError, 'The scan identification k'):
            self.sf[1.2]
        # non existant scan with "N.M" indexing
        with self.assertRaises(KeyError):
            self.sf["3.2"]

    def test_specfile_iterator(self):
        i = 0
        for scan in self.sf:
            if i == 1:
                self.assertEqual(scan.motor_positions,
                                 self.sf[1].motor_positions)
            i += 1
        # number of returned scans
        self.assertEqual(i, len(self.sf))

    def test_scan_index(self):
        self.assertEqual(self.scan1.index, 0)
        self.assertEqual(self.scan1_2.index, 3)
        self.assertEqual(self.scan25.index, 1)

    def test_scan_headers(self):
        self.assertEqual(self.scan25.scan_header_dict['S'],
                         "25  ascan  c3th 1.33245 1.52245  40 0.15")
        self.assertEqual(self.scan1.header[17], '#G0 0')
        self.assertEqual(len(self.scan1.header), 29)
        # parsing headers with long keys
        self.assertEqual(self.scan1.scan_header_dict['UMI0'],
                         'Current AutoM      Shutter')
        # parsing empty headers
        self.assertEqual(self.scan1.scan_header_dict['Q'], '')
        # duplicate headers: concatenated (with newline)
        self.assertEqual(self.scan1_2.scan_header_dict["U"],
                         "first duplicate line\nsecond duplicate line")

    def test_file_headers(self):
        self.assertEqual(self.scan1.header[1],
                         '#E 1455180875')
        self.assertEqual(self.scan1.file_header_dict['F'],
                         '/tmp/sf.dat')

    def test_multiple_file_headers(self):
        """Scan 1.2 is after the second file header, with a different
        Epoch"""
        self.assertEqual(self.scan1_2.header[1],
                         '#E 1455180876')

    def test_scan_labels(self):
        self.assertEqual(self.scan1.labels,
                         ['first column', 'second column', '3rd_col'])

    def test_data(self):
        # data_line() and data_col() take 1-based indices as arg
        self.assertAlmostEqual(self.scan1.data_line(1)[2],
                               1.56)
        # tests for data transposition between original file and .data attr
        self.assertAlmostEqual(self.scan1.data[2, 0],
                               8)
        self.assertEqual(self.scan1.data.shape, (3, 4))
        self.assertAlmostEqual(numpy.sum(self.scan1.data), 113.631)

    def test_data_column_by_name(self):
        self.assertAlmostEqual(self.scan25.data_column_by_name("col2")[1],
                               1.2)
        # Scan.data is transposed after readinq, so column is the first index
        self.assertAlmostEqual(numpy.sum(self.scan25.data_column_by_name("col2")),
                               numpy.sum(self.scan25.data[2, :]))
        with self.assertRaises(specfile.SfErrColNotFound):
            self.scan25.data_column_by_name("ygfxgfyxg")

    def test_motors(self):
        self.assertEqual(len(self.scan1.motor_names), 6)
        self.assertEqual(len(self.scan1.motor_positions), 6)
        self.assertAlmostEqual(sum(self.scan1.motor_positions),
                               223.385912)
        self.assertEqual(self.scan1.motor_names[1], 'MRTSlit UP')
        self.assertAlmostEqual(
            self.scan25.motor_position_by_name('MRTSlit UP'),
            -1.66875)

    def test_absence_of_file_header(self):
        """We expect Scan.file_header to be an empty list in the absence
        of a file header.
        """
        self.assertEqual(len(self.scan1_no_fhdr.motor_names), 0)
        # motor positions can still be read in the scan header
        # even in the absence of motor names
        self.assertAlmostEqual(sum(self.scan1_no_fhdr.motor_positions),
                               223.385912)
        self.assertEqual(len(self.scan1_no_fhdr.header), 15)
        self.assertEqual(len(self.scan1_no_fhdr.scan_header), 15)
        self.assertEqual(len(self.scan1_no_fhdr.file_header), 0)

    def test_crash_absence_of_file_header(self):
        """Test no crash in absence of file header and no leading newline
        character
        """
        self.assertEqual(len(self.scan1_no_fhdr_crash.motor_names), 0)
        # motor positions can still be read in the scan header
        # even in the absence of motor names
        self.assertAlmostEqual(sum(self.scan1_no_fhdr_crash.motor_positions),
                               223.385912)
        self.assertEqual(len(self.scan1_no_fhdr_crash.scan_header), 15)
        self.assertEqual(len(self.scan1_no_fhdr_crash.file_header), 0)

    def test_mca(self):
        self.assertEqual(len(self.scan1.mca), 0)
        self.assertEqual(len(self.scan1_2.mca), 3)
        self.assertEqual(self.scan1_2.mca[1][2], 5)
        self.assertEqual(sum(self.scan1_2.mca[2]), 21.7)

        # Negative indexing
        self.assertEqual(sum(self.scan1_2.mca[len(self.scan1_2.mca) - 1]),
                         sum(self.scan1_2.mca[-1]))

        # Test iterator
        line_count, total_sum = (0, 0)
        for mca_line in self.scan1_2.mca:
            line_count += 1
            total_sum += sum(mca_line)
        self.assertEqual(line_count, 3)
        self.assertAlmostEqual(total_sum, 36.8)

    def test_mca_header(self):
        self.assertEqual(self.scan1.mca_header_dict, {})
        self.assertEqual(len(self.scan1_2.mca_header_dict), 4)
        self.assertEqual(self.scan1_2.mca_header_dict["CALIB"], "1 2 3")
        self.assertEqual(self.scan1_2.mca.calibration,
                         [[1., 2., 3.]])
        # default calib in the absence of #@CALIB
        self.assertEqual(self.scan25.mca.calibration,
                         [[0., 1., 0.]])
        self.assertEqual(self.scan1_2.mca.channels,
                         [[0, 1, 2]])
        # absence of #@CHANN and spectra
        self.assertEqual(self.scan25.mca.channels,
                         [])

    @testutils.validate_logging(specfile._logger.name, warning=1)
    def test_empty_scan(self):
        """Test reading a scan with no data points"""
        self.assertEqual(len(self.empty_scan.labels),
                         3)
        col1 = self.empty_scan.data_column_by_name("second column")
        self.assertEqual(col1.shape, (0, ))


class TestSFLocale(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        fd, cls.fname = tempfile.mkstemp(text=False)
        if sys.version_info < (3, ):
            os.write(fd, sftext)
        else:
            os.write(fd, bytes(sftext, 'ascii'))
        os.close(fd)

    @classmethod
    def tearDownClass(cls):
        os.unlink(cls.fname)
        locale.setlocale(locale.LC_NUMERIC, loc)  # restore saved locale

    def crunch_data(self):
        self.sf3 = SpecFile(self.fname)
        self.assertAlmostEqual(self.sf3[0].data_line(1)[2],
                               1.56)
        self.sf3.close()

    @unittest.skipIf(not try_DE, "de_DE.utf8 locale not installed")
    def test_locale_de_DE(self):
        locale.setlocale(locale.LC_NUMERIC, 'de_DE.utf8')
        self.crunch_data()

    def test_locale_user(self):
        locale.setlocale(locale.LC_NUMERIC, '')  # use user's preferred locale
        self.crunch_data()

    def test_locale_C(self):
        locale.setlocale(locale.LC_NUMERIC, 'C')  # use default (C) locale
        self.crunch_data()
